# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Animals
"""

from wuttaweb.forms.schema import WuttaDictEnum

from wuttafarm.db.model import AnimalType, AnimalAsset
from wuttafarm.web.views.assets import AssetTypeMasterView, AssetMasterView
from wuttafarm.web.forms.schema import AnimalTypeRef
from wuttafarm.web.forms.widgets import ImageWidget


class AnimalTypeView(AssetTypeMasterView):
    """
    Master view for Animal Types
    """

    model_class = AnimalType
    route_prefix = "animal_types"
    url_prefix = "/animal-types"

    farmos_refurl_path = "/admin/structure/taxonomy/manage/animal_type/overview"

    grid_columns = [
        "name",
        "description",
        "changed",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "description",
        "changed",
        "farmos_uuid",
        "drupal_id",
    ]

    has_rows = True
    row_model_class = AnimalAsset
    rows_viewable = True

    row_grid_columns = [
        "asset_name",
        "sex",
        "is_sterile",
        "birthdate",
        "archived",
    ]

    rows_sort_defaults = "asset_name"

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_farmos_url(self, animal_type):
        return self.app.get_farmos_url(f"/taxonomy/term/{animal_type.drupal_id}")

    def get_xref_buttons(self, animal_type):
        buttons = super().get_xref_buttons(animal_type)

        if animal_type.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_animal_types.view", uuid=animal_type.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons

    def get_row_grid_data(self, animal_type):
        model = self.app.model
        session = self.Session()
        return (
            session.query(model.AnimalAsset)
            .join(model.Asset)
            .filter(model.AnimalAsset.animal_type == animal_type)
        )

    def configure_row_grid(self, grid):
        g = grid
        super().configure_row_grid(g)
        model = self.app.model
        enum = self.app.enum

        # asset_name
        g.set_link("asset_name")
        g.set_sorter("asset_name", model.Asset.asset_name)
        g.set_filter("asset_name", model.Asset.asset_name)

        # sex
        g.set_enum("sex", enum.ANIMAL_SEX)

        # archived
        g.set_renderer("archived", "boolean")
        g.set_sorter("archived", model.Asset.archived)
        g.set_filter("archived", model.Asset.archived)

    def get_row_action_url_view(self, animal, i):
        return self.request.route_url("animal_assets.view", uuid=animal.uuid)


class AnimalAssetView(AssetMasterView):
    """
    Master view for Animal Assets
    """

    model_class = AnimalAsset
    route_prefix = "animal_assets"
    url_prefix = "/assets/animal"

    farmos_refurl_path = "/assets/animal"

    labels = {
        "animal_type": "Species/Breed",
        "is_sterile": "Sterile",
    }

    grid_columns = [
        "thumbnail",
        "drupal_id",
        "asset_name",
        "animal_type",
        "birthdate",
        "is_sterile",
        "sex",
        "archived",
    ]

    form_fields = [
        "asset_name",
        "animal_type",
        "birthdate",
        "sex",
        "is_sterile",
        "notes",
        "asset_type",
        "archived",
        "farmos_uuid",
        "drupal_id",
        "thumbnail_url",
        "image_url",
        "thumbnail",
        "image",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        model = self.app.model
        enum = self.app.enum

        # animal_type
        g.set_joiner("animal_type", lambda q: q.join(model.AnimalType))
        g.set_sorter("animal_type", model.AnimalType.name)
        g.set_filter("animal_type", model.AnimalType.name)

        # birthdate
        g.set_renderer("birthdate", "date")

        # sex
        g.set_enum("sex", enum.ANIMAL_SEX)

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        enum = self.app.enum
        animal = f.model_instance

        # animal_type
        f.set_node("animal_type", AnimalTypeRef(self.request))

        # sex
        if self.viewing and animal.sex is None:
            pass  # TODO: dict enum widget does not handle null values well
        else:
            f.set_node("sex", WuttaDictEnum(self.request, enum.ANIMAL_SEX))
            f.set_required("sex", False)


def defaults(config, **kwargs):
    base = globals()

    AnimalTypeView = kwargs.get("AnimalTypeView", base["AnimalTypeView"])
    AnimalTypeView.defaults(config)

    AnimalAssetView = kwargs.get("AnimalAssetView", base["AnimalAssetView"])
    AnimalAssetView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
View for farmOS Users
"""

import datetime

import colander

from wuttaweb.forms.schema import WuttaDateTime

from wuttafarm.web.views.farmos import FarmOSMasterView


class UserView(FarmOSMasterView):
    """
    Master view for Farm Animals
    """

    model_name = "farmos_user"
    model_title = "farmOS User"
    model_title_plural = "farmOS Users"

    route_prefix = "farmos_users"
    url_prefix = "/farmOS/users"

    farmos_refurl_path = "/people"

    grid_columns = [
        "display_name",
    ]

    sort_defaults = "display_name"

    form_fields = [
        "display_name",
        "name",
        "mail",
        "timezone",
        "created",
        "changed",
    ]

    def get_grid_data(self, columns=None, session=None):
        users = self.farmos_client.resource.get("user", "user")
        return [self.normalize_user(u) for u in users["data"]]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # display_name
        g.set_link("display_name")
        g.set_searchable("display_name")

    def get_instance(self):
        user = self.farmos_client.resource.get_id(
            "user", "user", self.request.matchdict["uuid"]
        )
        self.raw_json = user
        return self.normalize_user(user["data"])

    def get_instance_title(self, user):
        return user["display_name"]

    def normalize_user(self, user):

        if created := user["attributes"].get("created"):
            created = datetime.datetime.fromisoformat(created)
            created = self.app.localtime(created)

        if changed := user["attributes"].get("changed"):
            changed = datetime.datetime.fromisoformat(changed)
            changed = self.app.localtime(changed)

        return {
            "uuid": user["id"],
            "drupal_id": user["attributes"].get("drupal_internal__uid"),
            "display_name": user["attributes"]["display_name"],
            "name": user["attributes"].get("name") or colander.null,
            "mail": user["attributes"].get("mail") or colander.null,
            "timezone": user["attributes"].get("timezone") or colander.null,
            "created": created,
            "changed": changed,
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        user = f.model_instance

        # created
        f.set_node("created", WuttaDateTime())

        # changed
        f.set_node("changed", WuttaDateTime())

    def get_xref_buttons(self, user):
        model = self.app.model
        session = self.Session()
        buttons = []

        if drupal_id := user["drupal_id"]:
            buttons.append(
                self.make_button(
                    "View in farmOS",
                    primary=True,
                    url=self.app.get_farmos_url(f"/user/{drupal_id}"),
                    target="_blank",
                    icon_left="external-link-alt",
                )
            )

        if wf_user := (
            session.query(model.WuttaFarmUser)
            .filter(model.WuttaFarmUser.farmos_uuid == user["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url("users.view", uuid=wf_user.uuid),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    UserView = kwargs.get("UserView", base["UserView"])
    UserView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base class for WuttaFarm master views
"""

from webhelpers2.html import tags

from wuttaweb.views import MasterView


class WuttaFarmMasterView(MasterView):
    """
    Base class for WuttaFarm master views
    """

    farmos_refurl_path = None

    labels = {
        "farmos_uuid": "farmOS UUID",
        "drupal_id": "Drupal ID",
        "image_url": "Image URL",
        "thumbnail_url": "Thumbnail URL",
    }

    row_labels = {
        "farmos_uuid": "farmOS UUID",
        "drupal_id": "Drupal ID",
        "image_url": "Image URL",
        "thumbnail_url": "Thumbnail URL",
    }

    def get_farmos_url(self, obj):
        return None

    def get_template_context(self, context):

        if self.listing and self.farmos_refurl_path:
            context["farmos_refurl"] = self.app.get_farmos_url(self.farmos_refurl_path)

        return context

    def render_grid_thumbnail(self, obj, field, value):
        if obj.thumbnail_url:
            return tags.image(
                obj.thumbnail_url, f"thumbnail for {self.get_model_title()}"
            )
        return None

    def get_xref_buttons(self, obj):
        url = self.get_farmos_url(obj)
        if url:
            return [
                self.make_button(
                    "View in farmOS",
                    primary=True,
                    url=url,
                    target="_blank",
                    icon_left="external-link-alt",
                )
            ]
        return []

    def configure_form(self, form):
        """ """
        f = form
        super().configure_form(f)

        # farmos_uuid
        if self.creating:
            f.remove("farmos_uuid")
        else:
            f.set_readonly("farmos_uuid")

        # drupal_id
        if self.creating:
            f.remove("drupal_id")
        else:
            f.set_readonly("drupal_id")

    def persist(self, obj, session=None):
        super().persist(obj, session)
        self.app.export_to_farmos(obj, require=False)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Model definition for Users (extension)
"""

import sqlalchemy as sa
from sqlalchemy import orm

from wuttjamaican.db import model


class WuttaFarmUser(model.Base):
    """
    WuttaFarm extension for the User model.
    """

    __tablename__ = "wuttafarm_user"
    __versioned__ = {}

    uuid = model.uuid_column(sa.ForeignKey("user.uuid"), default=None)

    user = orm.relationship(
        model.User,
        doc="""
        Reference to the User which this record extends.
        """,
        backref=orm.backref(
            "_wuttafarm",
            uselist=False,
            cascade="all, delete-orphan",
            cascade_backrefs=False,
            doc="""
            Reference to the WuttaFarm-specific extension record for
            the user.
            """,
        ),
    )

    farmos_uuid = sa.Column(
        model.UUID(),
        nullable=True,
        doc="""
        UUID for the user within farmOS
        """,
    )

    drupal_id = sa.Column(
        sa.Integer(),
        nullable=True,
        doc="""
        Drupal internal ID for the user.
        """,
    )

    def __str__(self):
        return str(self.user or "")


WuttaFarmUser.make_proxy(model.User, "_wuttafarm", "farmos_uuid")
WuttaFarmUser.make_proxy(model.User, "_wuttafarm", "drupal_id")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data import for farmOS -> WuttaFarm
"""

import datetime
import logging
from uuid import UUID

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

from wuttasync.importing import ImportHandler, ToWuttaHandler, Importer, ToWutta

from wuttafarm.db import model


log = logging.getLogger(__name__)


class FromFarmOSHandler(ImportHandler):
    """
    Base class for import handler using farmOS API as data source.
    """

    source_key = "farmos"
    generic_source_title = "farmOS"

    def begin_source_transaction(self):
        """
        Establish the farmOS API client.
        """
        token = self.get_farmos_oauth2_token()
        self.farmos_client = self.app.get_farmos_client(token=token)
        self.farmos_4x = self.app.is_farmos_4x(self.farmos_client)

    def get_farmos_oauth2_token(self):

        client_id = self.config.get(
            "farmos.oauth2.importing.client_id", default="wuttafarm"
        )
        client_secret = self.config.require("farmos.oauth2.importing.client_secret")
        scope = self.config.get("farmos.oauth2.importing.scope", default="farm_manager")

        client = BackendApplicationClient(client_id=client_id)
        oauth = OAuth2Session(client=client)

        return oauth.fetch_token(
            token_url=self.app.get_farmos_url("/oauth/token"),
            include_client_id=True,
            client_secret=client_secret,
            scope=scope,
        )

    def get_importer_kwargs(self, key, **kwargs):
        kwargs = super().get_importer_kwargs(key, **kwargs)
        kwargs["farmos_client"] = self.farmos_client
        kwargs["farmos_4x"] = self.farmos_4x
        return kwargs


class ToWuttaFarmHandler(ToWuttaHandler):
    """
    Base class for import handler targeting WuttaFarm
    """

    target_key = "wuttafarm"


class FromFarmOSToWuttaFarm(FromFarmOSHandler, ToWuttaFarmHandler):
    """
    Handler for farmOS → WuttaFarm import.
    """

    def define_importers(self):
        """ """
        importers = super().define_importers()
        importers["User"] = UserImporter
        importers["AssetType"] = AssetTypeImporter
        importers["LandType"] = LandTypeImporter
        importers["LandAsset"] = LandAssetImporter
        importers["StructureType"] = StructureTypeImporter
        importers["StructureAsset"] = StructureAssetImporter
        importers["AnimalType"] = AnimalTypeImporter
        importers["AnimalAsset"] = AnimalAssetImporter
        importers["GroupAsset"] = GroupAssetImporter
        importers["LogType"] = LogTypeImporter
        importers["ActivityLog"] = ActivityLogImporter
        return importers


class FromFarmOS(Importer):
    """
    Base class for importers using farmOS API as data source.
    """

    key = "farmos_uuid"

    def get_supported_fields(self):
        """
        Auto-remove the ``uuid`` field, since we use ``farmos_uuid``
        instead for the importer key.
        """
        fields = list(super().get_supported_fields())
        if "uuid" in fields:
            fields.remove("uuid")
        return fields

    def normalize_datetime(self, dt):
        """
        Convert a farmOS datetime value to naive UTC used by
        WuttaFarm.

        :param dt: Date/time string value "as-is" from the farmOS API.

        :returns: Equivalent naive UTC ``datetime``
        """
        dt = datetime.datetime.fromisoformat(dt)
        return self.app.make_utc(dt)


class ActivityLogImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Activity Logs
    """

    model_class = model.ActivityLog

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "message",
        "timestamp",
        "notes",
        "status",
    ]

    def get_source_objects(self):
        """ """
        logs = self.farmos_client.log.get("activity")
        return logs["data"]

    def normalize_source_object(self, log):
        """ """

        if notes := log["attributes"]["notes"]:
            notes = notes["value"]

        return {
            "farmos_uuid": UUID(log["id"]),
            "drupal_id": log["attributes"]["drupal_internal__id"],
            "message": log["attributes"]["name"],
            "timestamp": self.normalize_datetime(log["attributes"]["timestamp"]),
            "notes": notes,
            "status": log["attributes"]["status"],
        }


class AssetImporterBase(FromFarmOS, ToWutta):
    """
    Base class for farmOS API → WuttaFarm asset importers
    """

    def get_simple_fields(self):
        """ """
        fields = list(super().get_simple_fields())
        # nb. must explicitly declare proxy fields
        fields.extend(
            [
                "farmos_uuid",
                "drupal_id",
                "asset_type",
                "asset_name",
                "is_location",
                "is_fixed",
                "notes",
                "archived",
                "image_url",
                "thumbnail_url",
            ]
        )
        return fields

    def get_supported_fields(self):
        """ """
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "parents",
            ]
        )
        return fields

    def normalize_source_data(self, **kwargs):
        """ """
        data = super().normalize_source_data(**kwargs)

        if "parents" in self.fields:
            # nb. make sure parent-less (root) assets come first, so they
            # exist when child assets need to reference them
            data.sort(key=lambda l: len(l["parents"]))

        return data

    def normalize_asset(self, asset):
        """ """
        image_url = None
        thumbnail_url = None
        if relationships := asset.get("relationships"):

            if image := relationships.get("image"):
                if image["data"]:
                    image = self.farmos_client.resource.get_id(
                        "file", "file", image["data"][0]["id"]
                    )
                    if image_style := image["data"]["attributes"].get(
                        "image_style_uri"
                    ):
                        image_url = image_style["large"]
                        thumbnail_url = image_style["thumbnail"]

        if notes := asset["attributes"]["notes"]:
            notes = notes["value"]

        if self.farmos_4x:
            archived = asset["attributes"]["archived"]
        else:
            archived = asset["attributes"]["status"] == "archived"

        parents = None
        if "parents" in self.fields:
            parents = []
            for parent in asset["relationships"]["parent"]["data"]:
                parents.append((self.get_asset_type(parent), UUID(parent["id"])))

        return {
            "farmos_uuid": UUID(asset["id"]),
            "drupal_id": asset["attributes"]["drupal_internal__id"],
            "asset_name": asset["attributes"]["name"],
            "is_location": asset["attributes"]["is_location"],
            "is_fixed": asset["attributes"]["is_fixed"],
            "archived": archived,
            "notes": notes,
            "image_url": image_url,
            "thumbnail_url": thumbnail_url,
            "parents": parents,
        }

    def get_asset_type(self, asset):
        return asset["type"].split("--")[1]

    def normalize_target_object(self, asset):
        data = super().normalize_target_object(asset)

        if "parents" in self.fields:
            data["parents"] = [
                (p.parent.asset_type, p.parent.farmos_uuid)
                for p in asset.asset._parents
            ]

        return data

    def update_target_object(self, asset, source_data, target_data=None):
        model = self.app.model
        asset = super().update_target_object(asset, source_data, target_data)

        if "parents" in self.fields:
            if not target_data or target_data["parents"] != source_data["parents"]:

                for key in source_data["parents"]:
                    asset_type, farmos_uuid = key
                    if not target_data or key not in target_data["parents"]:
                        self.target_session.flush()
                        parent = (
                            self.target_session.query(model.Asset)
                            .filter(model.Asset.asset_type == asset_type)
                            .filter(model.Asset.farmos_uuid == farmos_uuid)
                            .one()
                        )
                        asset.asset._parents.append(model.AssetParent(parent=parent))

                if target_data:
                    for key in target_data["parents"]:
                        asset_type, farmos_uuid = key
                        if key not in source_data["parents"]:
                            parent = (
                                self.target_session.query(model.Asset)
                                .filter(model.Asset.asset_type == asset_type)
                                .filter(model.Asset.farmos_uuid == farmos_uuid)
                                .one()
                            )
                            parent = (
                                self.target_session.query(model.AssetParent)
                                .filter(model.AssetParent.asset == asset)
                                .filter(model.AssetParent.parent == parent)
                                .one()
                            )
                            self.target_session.delete(parent)

        return asset


class AnimalAssetImporter(AssetImporterBase):
    """
    farmOS API → WuttaFarm importer for Animals
    """

    model_class = model.AnimalAsset

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "asset_type",
        "asset_name",
        "animal_type_uuid",
        "sex",
        "is_sterile",
        "birthdate",
        "notes",
        "archived",
        "image_url",
        "thumbnail_url",
    ]

    def setup(self):
        super().setup()
        model = self.app.model

        self.animal_types_by_farmos_uuid = {}
        for animal_type in self.target_session.query(model.AnimalType):
            if animal_type.farmos_uuid:
                self.animal_types_by_farmos_uuid[animal_type.farmos_uuid] = animal_type

    def get_source_objects(self):
        """ """
        animals = self.farmos_client.asset.get("animal")
        return animals["data"]

    def normalize_source_object(self, animal):
        """ """
        animal_type_uuid = None
        if relationships := animal.get("relationships"):

            if animal_type := relationships.get("animal_type"):
                if animal_type["data"]:
                    if wf_animal_type := self.animal_types_by_farmos_uuid.get(
                        UUID(animal_type["data"]["id"])
                    ):
                        animal_type_uuid = wf_animal_type.uuid
                    else:
                        log.warning(
                            "animal type not found: %s", animal_type["data"]["id"]
                        )

        if not animal_type_uuid:
            log.warning("missing/invalid animal_type for farmOS Animal: %s", animal)
            return None

        birthdate = animal["attributes"]["birthdate"]
        if birthdate:
            birthdate = datetime.datetime.fromisoformat(birthdate)
            birthdate = self.app.localtime(birthdate)
            birthdate = self.app.make_utc(birthdate)

        if self.farmos_4x:
            sterile = animal["attributes"]["is_sterile"]
        else:
            sterile = animal["attributes"]["is_castrated"]

        data = self.normalize_asset(animal)
        data.update(
            {
                "asset_type": "animal",
                "animal_type_uuid": animal_type_uuid,
                "sex": animal["attributes"]["sex"],
                "is_sterile": sterile,
                "birthdate": birthdate,
            }
        )
        return data


class AnimalTypeImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Animal Types
    """

    model_class = model.AnimalType

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
        "description",
        "changed",
    ]

    def get_source_objects(self):
        """ """
        animal_types = self.farmos_client.resource.get("taxonomy_term", "animal_type")
        return animal_types["data"]

    def normalize_source_object(self, animal_type):
        """ """
        return {
            "farmos_uuid": UUID(animal_type["id"]),
            "drupal_id": animal_type["attributes"]["drupal_internal__tid"],
            "name": animal_type["attributes"]["name"],
            "description": animal_type["attributes"]["description"],
            "changed": self.normalize_datetime(animal_type["attributes"]["changed"]),
        }


class AssetTypeImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Asset Types
    """

    model_class = model.AssetType

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
        "description",
    ]

    def get_source_objects(self):
        """ """
        asset_types = self.farmos_client.resource.get("asset_type")
        return asset_types["data"]

    def normalize_source_object(self, asset_type):
        """ """
        return {
            "farmos_uuid": UUID(asset_type["id"]),
            "drupal_id": asset_type["attributes"]["drupal_internal__id"],
            "name": asset_type["attributes"]["label"],
            "description": asset_type["attributes"]["description"],
        }


class GroupAssetImporter(AssetImporterBase):
    """
    farmOS API → WuttaFarm importer for Group Assets
    """

    model_class = model.GroupAsset

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "asset_type",
        "asset_name",
        "is_location",
        "is_fixed",
        "notes",
        "archived",
        "image_url",
        "thumbnail_url",
        "parents",
    ]

    def get_source_objects(self):
        """ """
        groups = self.farmos_client.asset.get("group")
        return groups["data"]

    def normalize_source_object(self, group):
        """ """
        data = self.normalize_asset(group)
        data.update(
            {
                "asset_type": "group",
            }
        )
        return data


class LandAssetImporter(AssetImporterBase):
    """
    farmOS API → WuttaFarm importer for Land Assets
    """

    model_class = model.LandAsset

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "asset_type",
        "asset_name",
        "land_type_uuid",
        "is_location",
        "is_fixed",
        "notes",
        "archived",
        "parents",
    ]

    def setup(self):
        """ """
        super().setup()
        model = self.app.model

        self.land_types_by_id = {}
        for land_type in self.target_session.query(model.LandType):
            self.land_types_by_id[land_type.drupal_id] = land_type

    def get_source_objects(self):
        """ """
        land_assets = self.farmos_client.asset.get("land")
        return land_assets["data"]

    def normalize_source_object(self, land):
        """ """
        land_type_id = land["attributes"]["land_type"]
        land_type = self.land_types_by_id.get(land_type_id)
        if not land_type:
            log.warning(
                "invalid land_type '%s' for farmOS Land Asset: %s", land_type_id, land
            )
            return None

        data = self.normalize_asset(land)
        data.update(
            {
                "asset_type": "land",
                "land_type_uuid": land_type.uuid,
            }
        )
        return data


class LandTypeImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Land Types
    """

    model_class = model.LandType

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
    ]

    def get_source_objects(self):
        """ """
        land_types = self.farmos_client.resource.get("land_type")
        return land_types["data"]

    def normalize_source_object(self, land_type):
        """ """
        return {
            "farmos_uuid": UUID(land_type["id"]),
            "drupal_id": land_type["attributes"]["drupal_internal__id"],
            "name": land_type["attributes"]["label"],
        }


class LogTypeImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Log Types
    """

    model_class = model.LogType

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
        "description",
    ]

    def get_source_objects(self):
        """ """
        log_types = self.farmos_client.resource.get("log_type")
        return log_types["data"]

    def normalize_source_object(self, log_type):
        """ """
        return {
            "farmos_uuid": UUID(log_type["id"]),
            "drupal_id": log_type["attributes"]["drupal_internal__id"],
            "name": log_type["attributes"]["label"],
            "description": log_type["attributes"]["description"],
        }


class StructureAssetImporter(AssetImporterBase):
    """
    farmOS API → WuttaFarm importer for Structure Assets
    """

    model_class = model.StructureAsset

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "asset_type",
        "asset_name",
        "structure_type_uuid",
        "is_location",
        "is_fixed",
        "notes",
        "archived",
        "image_url",
        "thumbnail_url",
        "parents",
    ]

    def setup(self):
        super().setup()
        model = self.app.model

        self.structure_types_by_id = {}
        for structure_type in self.target_session.query(model.StructureType):
            self.structure_types_by_id[structure_type.drupal_id] = structure_type

    def get_source_objects(self):
        """ """
        structures = self.farmos_client.asset.get("structure")
        return structures["data"]

    def normalize_source_object(self, structure):
        """ """
        structure_type_id = structure["attributes"]["structure_type"]
        structure_type = self.structure_types_by_id.get(structure_type_id)
        if not structure_type:
            log.warning(
                "invalid structure_type '%s' for farmOS Structure Asset: %s",
                structure_type_id,
                structure,
            )
            return None

        data = self.normalize_asset(structure)
        data.update(
            {
                "asset_type": "structure",
                "structure_type_uuid": structure_type.uuid,
            }
        )
        return data


class StructureTypeImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Structure Types
    """

    model_class = model.StructureType

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
    ]

    def get_source_objects(self):
        """ """
        structure_types = self.farmos_client.resource.get("structure_type")
        return structure_types["data"]

    def normalize_source_object(self, structure_type):
        """ """
        return {
            "farmos_uuid": UUID(structure_type["id"]),
            "drupal_id": structure_type["attributes"]["drupal_internal__id"],
            "name": structure_type["attributes"]["label"],
        }


class UserImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Users
    """

    model_class = model.User

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "username",
    ]

    def get_simple_fields(self):
        """ """
        fields = list(super().get_simple_fields())
        # nb. must explicitly declare extension fields
        fields.extend(
            [
                "farmos_uuid",
                "drupal_id",
            ]
        )
        return fields

    def get_source_objects(self):
        """ """
        users = self.farmos_client.resource.get("user")
        return users["data"]

    def normalize_source_object(self, user):
        """ """

        # nb. skip Anonymous user which does not have drupal id
        drupal_id = user["attributes"].get("drupal_internal__uid")
        if not drupal_id:
            return None

        return {
            "farmos_uuid": UUID(user["id"]),
            "drupal_id": drupal_id,
            "username": user["attributes"]["name"],
        }

    def can_delete_object(self, user, data=None):
        """
        Prevent delete for users which do not exist in farmOS.
        """
        if not user.farmos_uuid:
            return False
        return True



"""add generic, animal assets

Revision ID: d6e8d16d6854
Revises: 554e6168c339
Create Date: 2026-02-15 09:11:04.886362

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "d6e8d16d6854"
down_revision: Union[str, None] = "554e6168c339"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # animal
    op.drop_table("animal")
    op.drop_index(
        op.f("ix_animal_version_end_transaction_id"), table_name="animal_version"
    )
    op.drop_index(op.f("ix_animal_version_operation_type"), table_name="animal_version")
    op.drop_index(
        op.f("ix_animal_version_pk_transaction_id"), table_name="animal_version"
    )
    op.drop_index(op.f("ix_animal_version_pk_validity"), table_name="animal_version")
    op.drop_index(op.f("ix_animal_version_transaction_id"), table_name="animal_version")
    op.drop_table("animal_version")

    # asset
    op.create_table(
        "asset",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("farmos_uuid", wuttjamaican.db.util.UUID(), nullable=True),
        sa.Column("drupal_id", sa.Integer(), nullable=True),
        sa.Column("asset_type", sa.String(length=100), nullable=False),
        sa.Column("asset_name", sa.String(length=100), nullable=False),
        sa.Column("is_location", sa.Boolean(), nullable=False),
        sa.Column("is_fixed", sa.Boolean(), nullable=False),
        sa.Column("notes", sa.Text(), nullable=True),
        sa.Column("thumbnail_url", sa.String(length=255), nullable=True),
        sa.Column("image_url", sa.String(length=255), nullable=True),
        sa.Column("archived", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["asset_type"],
            ["asset_type.drupal_id"],
            name=op.f("fk_asset_asset_type_asset_type"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_asset")),
        sa.UniqueConstraint("drupal_id", name=op.f("uq_asset_drupal_id")),
        sa.UniqueConstraint("farmos_uuid", name=op.f("uq_asset_farmos_uuid")),
    )
    op.create_table(
        "asset_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "farmos_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("drupal_id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column(
            "asset_type", sa.String(length=100), autoincrement=False, nullable=True
        ),
        sa.Column(
            "asset_name", sa.String(length=100), autoincrement=False, nullable=True
        ),
        sa.Column("is_location", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column("is_fixed", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column("notes", sa.Text(), autoincrement=False, nullable=True),
        sa.Column(
            "thumbnail_url", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "image_url", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column("archived", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_asset_version")
        ),
    )
    op.create_index(
        op.f("ix_asset_version_end_transaction_id"),
        "asset_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_version_operation_type"),
        "asset_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_asset_version_pk_transaction_id",
        "asset_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_asset_version_pk_validity",
        "asset_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_version_transaction_id"),
        "asset_version",
        ["transaction_id"],
        unique=False,
    )

    # asset_animal
    op.create_table(
        "asset_animal",
        sa.Column("animal_type_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("birthdate", sa.DateTime(), nullable=True),
        sa.Column("sex", sa.String(length=1), nullable=True),
        sa.Column("is_sterile", sa.Boolean(), nullable=True),
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["animal_type_uuid"],
            ["animal_type.uuid"],
            name=op.f("fk_asset_animal_animal_type_uuid_animal_type"),
        ),
        sa.ForeignKeyConstraint(
            ["uuid"], ["asset.uuid"], name=op.f("fk_asset_animal_uuid_asset")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_asset_animal")),
    )
    op.create_table(
        "asset_animal_version",
        sa.Column(
            "animal_type_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("birthdate", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column("sex", sa.String(length=1), autoincrement=False, nullable=True),
        sa.Column("is_sterile", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_asset_animal_version")
        ),
    )
    op.create_index(
        op.f("ix_asset_animal_version_end_transaction_id"),
        "asset_animal_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_animal_version_operation_type"),
        "asset_animal_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_asset_animal_version_pk_transaction_id",
        "asset_animal_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_asset_animal_version_pk_validity",
        "asset_animal_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_animal_version_transaction_id"),
        "asset_animal_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # asset_animal
    op.drop_index(
        op.f("ix_asset_animal_version_transaction_id"),
        table_name="asset_animal_version",
    )
    op.drop_index(
        "ix_asset_animal_version_pk_validity", table_name="asset_animal_version"
    )
    op.drop_index(
        "ix_asset_animal_version_pk_transaction_id", table_name="asset_animal_version"
    )
    op.drop_index(
        op.f("ix_asset_animal_version_operation_type"),
        table_name="asset_animal_version",
    )
    op.drop_index(
        op.f("ix_asset_animal_version_end_transaction_id"),
        table_name="asset_animal_version",
    )
    op.drop_table("asset_animal_version")
    op.drop_table("asset_animal")

    # asset
    op.drop_index(op.f("ix_asset_version_transaction_id"), table_name="asset_version")
    op.drop_index("ix_asset_version_pk_validity", table_name="asset_version")
    op.drop_index("ix_asset_version_pk_transaction_id", table_name="asset_version")
    op.drop_index(op.f("ix_asset_version_operation_type"), table_name="asset_version")
    op.drop_index(
        op.f("ix_asset_version_end_transaction_id"), table_name="asset_version"
    )
    op.drop_table("asset_version")
    op.drop_table("asset")

    # animal
    op.create_table(
        "animal",
        sa.Column("uuid", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column("name", sa.VARCHAR(length=100), autoincrement=False, nullable=False),
        sa.Column("animal_type_uuid", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column("birthdate", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column("sex", sa.VARCHAR(length=1), autoincrement=False, nullable=True),
        sa.Column("is_sterile", sa.BOOLEAN(), autoincrement=False, nullable=True),
        sa.Column("archived", sa.BOOLEAN(), autoincrement=False, nullable=False),
        sa.Column("notes", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column(
            "image_url", sa.VARCHAR(length=255), autoincrement=False, nullable=True
        ),
        sa.Column("farmos_uuid", sa.UUID(), autoincrement=False, nullable=True),
        sa.Column("drupal_id", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column(
            "thumbnail_url", sa.VARCHAR(length=255), autoincrement=False, nullable=True
        ),
        sa.ForeignKeyConstraint(
            ["animal_type_uuid"],
            ["animal_type.uuid"],
            name=op.f("fk_animal_animal_type_uuid_animal_type"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_animal")),
        sa.UniqueConstraint(
            "drupal_id",
            name=op.f("uq_animal_drupal_id"),
            postgresql_include=[],
            postgresql_nulls_not_distinct=False,
        ),
        sa.UniqueConstraint(
            "farmos_uuid",
            name=op.f("uq_animal_farmos_uuid"),
            postgresql_include=[],
            postgresql_nulls_not_distinct=False,
        ),
    )
    op.create_table(
        "animal_version",
        sa.Column("uuid", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column("name", sa.VARCHAR(length=100), autoincrement=False, nullable=True),
        sa.Column("animal_type_uuid", sa.UUID(), autoincrement=False, nullable=True),
        sa.Column(
            "birthdate", postgresql.TIMESTAMP(), autoincrement=False, nullable=True
        ),
        sa.Column("sex", sa.VARCHAR(length=1), autoincrement=False, nullable=True),
        sa.Column("is_sterile", sa.BOOLEAN(), autoincrement=False, nullable=True),
        sa.Column("archived", sa.BOOLEAN(), autoincrement=False, nullable=True),
        sa.Column("notes", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column(
            "image_url", sa.VARCHAR(length=255), autoincrement=False, nullable=True
        ),
        sa.Column("farmos_uuid", sa.UUID(), autoincrement=False, nullable=True),
        sa.Column("drupal_id", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column("transaction_id", sa.BIGINT(), autoincrement=False, nullable=False),
        sa.Column(
            "end_transaction_id", sa.BIGINT(), autoincrement=False, nullable=True
        ),
        sa.Column("operation_type", sa.SMALLINT(), autoincrement=False, nullable=False),
        sa.Column(
            "thumbnail_url", sa.VARCHAR(length=255), autoincrement=False, nullable=True
        ),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_animal_version")
        ),
    )
    op.create_index(
        op.f("ix_animal_version_transaction_id"),
        "animal_version",
        ["transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_animal_version_pk_validity"),
        "animal_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_animal_version_pk_transaction_id"),
        "animal_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        op.f("ix_animal_version_operation_type"),
        "animal_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_animal_version_end_transaction_id"),
        "animal_version",
        ["end_transaction_id"],
        unique=False,
    )



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Farm Animals
"""

import datetime

import colander

from wuttaweb.forms.schema import WuttaDateTime
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views.farmos import FarmOSMasterView
from wuttafarm.web.forms.schema import UsersType, AnimalTypeType, StructureType
from wuttafarm.web.forms.widgets import ImageWidget


class AnimalView(FarmOSMasterView):
    """
    Master view for Farm Animals
    """

    model_name = "farmos_animal"
    model_title = "farmOS Animal"
    model_title_plural = "farmOS Animals"

    route_prefix = "farmos_animals"
    url_prefix = "/farmOS/animals"

    farmos_refurl_path = "/assets/animal"

    labels = {
        "animal_type": "Species / Breed",
        "location": "Current Location",
    }

    grid_columns = [
        "name",
        "birthdate",
        "sex",
        "is_sterile",
        "archived",
    ]

    sort_defaults = "name"

    form_fields = [
        "name",
        "animal_type",
        "birthdate",
        "sex",
        "is_sterile",
        "archived",
        "owners",
        "location",
        "notes",
        "raw_image_url",
        "large_image_url",
        "thumbnail_image_url",
        "image",
    ]

    def get_grid_data(self, columns=None, session=None):
        animals = self.farmos_client.resource.get("asset", "animal")
        return [self.normalize_animal(a) for a in animals["data"]]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")
        g.set_searchable("name")

        # birthdate
        g.set_renderer("birthdate", "date")

        # is_sterile
        g.set_renderer("is_sterile", "boolean")

        # archived
        g.set_renderer("archived", "boolean")

    def get_instance(self):

        animal = self.farmos_client.resource.get_id(
            "asset", "animal", self.request.matchdict["uuid"]
        )
        self.raw_json = animal

        # instance data
        data = self.normalize_animal(animal["data"])

        if relationships := animal["data"].get("relationships"):

            # add animal type
            if animal_type := relationships.get("animal_type"):
                if animal_type["data"]:
                    animal_type = self.farmos_client.resource.get_id(
                        "taxonomy_term", "animal_type", animal_type["data"]["id"]
                    )
                    data["animal_type"] = {
                        "uuid": animal_type["data"]["id"],
                        "name": animal_type["data"]["attributes"]["name"],
                    }

            # add location
            if location := relationships.get("location"):
                if location["data"]:
                    location = self.farmos_client.resource.get_id(
                        "asset", "structure", location["data"][0]["id"]
                    )
                    data["location"] = {
                        "uuid": location["data"]["id"],
                        "name": location["data"]["attributes"]["name"],
                    }

            # add owners
            if owner := relationships.get("owner"):
                data["owners"] = []
                for owner_data in owner["data"]:
                    owner = self.farmos_client.resource.get_id(
                        "user", "user", owner_data["id"]
                    )
                    data["owners"].append(
                        {
                            "uuid": owner["data"]["id"],
                            "display_name": owner["data"]["attributes"]["display_name"],
                        }
                    )

            # add image urls
            if image := relationships.get("image"):
                if image["data"]:
                    image = self.farmos_client.resource.get_id(
                        "file", "file", image["data"][0]["id"]
                    )
                    data["raw_image_url"] = self.app.get_farmos_url(
                        image["data"]["attributes"]["uri"]["url"]
                    )
                    # nb. other styles available:  medium, wide
                    data["large_image_url"] = image["data"]["attributes"][
                        "image_style_uri"
                    ]["large"]
                    data["thumbnail_image_url"] = image["data"]["attributes"][
                        "image_style_uri"
                    ]["thumbnail"]

        return data

    def get_instance_title(self, animal):
        return animal["name"]

    def normalize_animal(self, animal):

        birthdate = animal["attributes"]["birthdate"]
        if birthdate:
            birthdate = datetime.datetime.fromisoformat(birthdate)
            birthdate = self.app.localtime(birthdate)

        sterile = None
        if self.farmos_4x:
            sterile = animal["attributes"]["is_sterile"]
        else:
            sterile = animal["attributes"]["is_castrated"]

        if notes := animal["attributes"]["notes"]:
            notes = notes["value"]

        if self.farmos_4x:
            archived = animal["attributes"]["archived"]
        else:
            archived = animal["attributes"]["status"] == "archived"

        return {
            "uuid": animal["id"],
            "drupal_id": animal["attributes"]["drupal_internal__id"],
            "name": animal["attributes"]["name"],
            "birthdate": birthdate,
            "sex": animal["attributes"]["sex"] or colander.null,
            "is_sterile": sterile,
            "location": colander.null,  # TODO
            "archived": archived,
            "notes": notes or colander.null,
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        animal = f.model_instance

        # animal_type
        f.set_node("animal_type", AnimalTypeType(self.request))

        # birthdate
        f.set_node("birthdate", WuttaDateTime())
        f.set_widget("birthdate", WuttaDateTimeWidget(self.request))

        # is_sterile
        f.set_node("is_sterile", colander.Boolean())

        # location
        f.set_node("location", StructureType(self.request))

        # owners
        f.set_node("owners", UsersType(self.request))

        # notes
        f.set_widget("notes", "notes")

        # archived
        f.set_node("archived", colander.Boolean())

        # image
        if url := animal.get("large_image_url"):
            f.set_widget("image", ImageWidget("animal image"))
            f.set_default("image", url)

    def get_xref_buttons(self, animal):
        model = self.app.model
        session = self.Session()

        buttons = [
            self.make_button(
                "View in farmOS",
                primary=True,
                url=self.app.get_farmos_url(f"/asset/{animal['drupal_id']}"),
                target="_blank",
                icon_left="external-link-alt",
            ),
        ]

        if wf_animal := (
            session.query(model.Asset)
            .filter(model.Asset.farmos_uuid == animal["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "animal_assets.view", uuid=wf_animal.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    AnimalView = kwargs.get("AnimalView", base["AnimalView"])
    AnimalView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Custom form widgets for WuttaFarm
"""

import json

import colander

from wuttaweb.forms.schema import ObjectRef, WuttaSet


class AnimalTypeRef(ObjectRef):
    """
    Custom schema type for a
    :class:`~wuttafarm.db.model.animals.AnimalType` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        return model.AnimalType

    def sort_query(self, query):  # pylint: disable=empty-docstring
        """ """
        return query.order_by(self.model_class.name)

    def get_object_url(self, obj):  # pylint: disable=empty-docstring
        """ """
        animal_type = obj
        return self.request.route_url("animal_types.view", uuid=animal_type.uuid)


class AnimalTypeType(colander.SchemaType):

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return json.dumps(appstruct)

    def widget_maker(self, **kwargs):  # pylint: disable=empty-docstring
        """ """
        from wuttafarm.web.forms.widgets import AnimalTypeWidget

        return AnimalTypeWidget(self.request, **kwargs)


class LandTypeRef(ObjectRef):
    """
    Custom schema type for a
    :class:`~wuttafarm.db.model.land.LandType` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        return model.LandType

    def sort_query(self, query):  # pylint: disable=empty-docstring
        """ """
        return query.order_by(self.model_class.name)

    def get_object_url(self, obj):  # pylint: disable=empty-docstring
        """ """
        land_type = obj
        return self.request.route_url("land_types.view", uuid=land_type.uuid)


class StructureType(colander.SchemaType):

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return json.dumps(appstruct)

    def widget_maker(self, **kwargs):  # pylint: disable=empty-docstring
        """ """
        from wuttafarm.web.forms.widgets import StructureWidget

        return StructureWidget(self.request, **kwargs)


class StructureTypeRef(ObjectRef):
    """
    Custom schema type for a
    :class:`~wuttafarm.db.model.structures.Structure` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        return model.StructureType

    def sort_query(self, query):  # pylint: disable=empty-docstring
        """ """
        return query.order_by(self.model_class.name)

    def get_object_url(self, obj):  # pylint: disable=empty-docstring
        """ """
        structure_type = obj
        return self.request.route_url("structure_types.view", uuid=structure_type.uuid)


class UsersType(colander.SchemaType):

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return json.dumps(appstruct)

    def widget_maker(self, **kwargs):  # pylint: disable=empty-docstring
        """ """
        from wuttafarm.web.forms.widgets import UsersWidget

        return UsersWidget(self.request, **kwargs)


class AssetParentRefs(WuttaSet):
    """
    Schema type for Parents field which references assets.
    """

    def serialize(self, node, appstruct):
        if not appstruct:
            appstruct = []
        uuids = [u.hex for u in appstruct]
        return json.dumps(uuids)

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import AssetParentRefsWidget

        return AssetParentRefsWidget(self.request, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaFarm Menu
"""

from wuttaweb import menus as base


class WuttaFarmMenuHandler(base.MenuHandler):
    """
    WuttaFarm menu handler
    """

    def make_menus(self, request, **kwargs):
        return [
            self.make_asset_menu(request),
            self.make_log_menu(request),
            self.make_farmos_menu(request),
            self.make_admin_menu(request, include_people=True),
        ]

    def make_asset_menu(self, request):
        return {
            "title": "Assets",
            "type": "menu",
            "items": [
                {
                    "title": "All Assets",
                    "route": "assets",
                    "perm": "assets.list",
                },
                {
                    "title": "Animal",
                    "route": "animal_assets",
                    "perm": "animal_assets.list",
                },
                {
                    "title": "Group",
                    "route": "group_assets",
                    "perm": "group_assets.list",
                },
                {
                    "title": "Land",
                    "route": "land_assets",
                    "perm": "land_assets.list",
                },
                {
                    "title": "Structure",
                    "route": "structure_assets",
                    "perm": "structure_assets.list",
                },
                {"type": "sep"},
                {
                    "title": "Animal Types",
                    "route": "animal_types",
                    "perm": "animal_types.list",
                },
                {
                    "title": "Land Types",
                    "route": "land_types",
                    "perm": "land_types.list",
                },
                {
                    "title": "Structure Types",
                    "route": "structure_types",
                    "perm": "structure_types.list",
                },
                {
                    "title": "Asset Types",
                    "route": "asset_types",
                    "perm": "asset_types.list",
                },
            ],
        }

    def make_log_menu(self, request):
        return {
            "title": "Logs",
            "type": "menu",
            "items": [
                {
                    "title": "Activity Logs",
                    "route": "activity_logs",
                    "perm": "activity_logs.list",
                },
                {"type": "sep"},
                {
                    "title": "Log Types",
                    "route": "log_types",
                    "perm": "log_types.list",
                },
            ],
        }

    def make_farmos_menu(self, request):
        config = request.wutta_config
        app = config.get_app()
        return {
            "title": "farmOS",
            "type": "menu",
            "items": [
                {
                    "title": "Go to farmOS",
                    "url": app.get_farmos_url(),
                    "target": "_blank",
                },
                {"type": "sep"},
                {
                    "title": "Animals",
                    "route": "farmos_animals",
                    "perm": "farmos_animals.list",
                },
                {
                    "title": "Groups",
                    "route": "farmos_groups",
                    "perm": "farmos_groups.list",
                },
                {
                    "title": "Structures",
                    "route": "farmos_structures",
                    "perm": "farmos_structures.list",
                },
                {
                    "title": "Land",
                    "route": "farmos_land_assets",
                    "perm": "farmos_land_assets.list",
                },
                {"type": "sep"},
                {
                    "title": "Activity Logs",
                    "route": "farmos_logs_activity",
                    "perm": "farmos_logs_activity.list",
                },
                {"type": "sep"},
                {
                    "title": "Animal Types",
                    "route": "farmos_animal_types",
                    "perm": "farmos_animal_types.list",
                },
                {
                    "title": "Structure Types",
                    "route": "farmos_structure_types",
                    "perm": "farmos_structure_types.list",
                },
                {
                    "title": "Land Types",
                    "route": "farmos_land_types",
                    "perm": "farmos_land_types.list",
                },
                {
                    "title": "Asset Types",
                    "route": "farmos_asset_types",
                    "perm": "farmos_asset_types.list",
                },
                {
                    "title": "Log Types",
                    "route": "farmos_log_types",
                    "perm": "farmos_log_types.list",
                },
                {"type": "sep"},
                {
                    "title": "Users",
                    "route": "farmos_users",
                    "perm": "farmos_users.list",
                },
            ],
        }



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Importer models targeting farmOS
"""

import datetime
from uuid import UUID

import requests

from wuttasync.importing import Importer


class ToFarmOS(Importer):
    """
    Base class for data importer targeting the farmOS API.
    """

    key = "uuid"
    caches_target = True

    def format_datetime(self, dt):
        """
        Convert a WuttaFarm datetime object to the format required for
        pushing to the farmOS API.
        """
        if dt is None:
            return None
        dt = self.app.localtime(dt)
        return dt.timestamp()

    def normalize_datetime(self, dt):
        """
        Convert a farmOS datetime value to naive UTC used by
        WuttaFarm.

        :param dt: Date/time string value "as-is" from the farmOS API.

        :returns: Equivalent naive UTC ``datetime``
        """
        if dt is None:
            return None
        dt = datetime.datetime.fromisoformat(dt)
        return self.app.make_utc(dt)


class ToFarmOSAsset(ToFarmOS):
    """
    Base class for asset data importer targeting the farmOS API.
    """

    farmos_asset_type = None

    def get_target_objects(self, **kwargs):
        assets = self.farmos_client.asset.get(self.farmos_asset_type)
        return assets["data"]

    def get_target_object(self, key):

        # fetch from cache, if applicable
        if self.caches_target:
            return super().get_target_object(key)

        # okay now must fetch via API
        if self.get_keys() != ["uuid"]:
            raise ValueError("must use uuid key for this to work")
        uuid = key[0]

        try:
            asset = self.farmos_client.asset.get_id(self.farmos_asset_type, str(uuid))
        except requests.HTTPError as exc:
            if exc.response.status_code == 404:
                return None
            raise
        return asset["data"]

    def create_target_object(self, key, source_data):
        if source_data.get("__ignoreme__"):
            return None
        if self.dry_run:
            return source_data

        payload = self.get_asset_payload(source_data)
        result = self.farmos_client.asset.send(self.farmos_asset_type, payload)
        normal = self.normalize_target_object(result["data"])
        normal["_new_object"] = result["data"]
        return normal

    def update_target_object(self, asset, source_data, target_data=None):
        if self.dry_run:
            return asset

        payload = self.get_asset_payload(source_data)
        payload["id"] = str(source_data["uuid"])
        result = self.farmos_client.asset.send(self.farmos_asset_type, payload)
        return self.normalize_target_object(result["data"])

    def normalize_target_object(self, asset):

        if notes := asset["attributes"]["notes"]:
            notes = notes["value"]

        return {
            "uuid": UUID(asset["id"]),
            "asset_name": asset["attributes"]["name"],
            "is_location": asset["attributes"]["is_location"],
            "is_fixed": asset["attributes"]["is_fixed"],
            "notes": notes,
            "archived": asset["attributes"]["archived"],
        }

    def get_asset_payload(self, source_data):

        attrs = {}
        if "asset_name" in self.fields:
            attrs["name"] = source_data["asset_name"]
        if "is_location" in self.fields:
            attrs["is_location"] = source_data["is_location"]
        if "is_fixed" in self.fields:
            attrs["is_fixed"] = source_data["is_fixed"]
        if "notes" in self.fields:
            attrs["notes"] = {"value": source_data["notes"]}
        if "archived" in self.fields:
            attrs["archived"] = source_data["archived"]

        payload = {"attributes": attrs}

        return payload


class AnimalAssetImporter(ToFarmOSAsset):

    model_title = "AnimalAsset"
    farmos_asset_type = "animal"

    supported_fields = [
        "uuid",
        "asset_name",
        "animal_type_uuid",
        "sex",
        "is_sterile",
        "birthdate",
        "notes",
        "archived",
    ]

    def normalize_target_object(self, animal):
        data = super().normalize_target_object(animal)
        data.update(
            {
                "animal_type_uuid": UUID(
                    animal["relationships"]["animal_type"]["data"]["id"]
                ),
                "sex": animal["attributes"]["sex"],
                "is_sterile": animal["attributes"]["is_sterile"],
                "birthdate": self.normalize_datetime(animal["attributes"]["birthdate"]),
            }
        )
        return data

    def get_asset_payload(self, source_data):
        payload = super().get_asset_payload(source_data)

        attrs = {}
        if "sex" in self.fields:
            attrs["sex"] = source_data["sex"]
        if "is_sterile" in self.fields:
            attrs["is_sterile"] = source_data["is_sterile"]
        if "birthdate" in self.fields:
            attrs["birthdate"] = self.format_datetime(source_data["birthdate"])

        rels = {}
        if "animal_type_uuid" in self.fields:
            rels["animal_type"] = {
                "data": {
                    "id": str(source_data["animal_type_uuid"]),
                    "type": "taxonomy_term--animal_type",
                }
            }

        payload["attributes"].update(attrs)
        if rels:
            payload.setdefault("relationships", {}).update(rels)

        return payload


class AnimalTypeImporter(ToFarmOS):

    model_title = "AnimalType"

    supported_fields = [
        "uuid",
        "name",
    ]

    def get_target_objects(self, **kwargs):
        result = self.farmos_client.resource.get("taxonomy_term", "animal_type")
        return result["data"]

    def get_target_object(self, key):

        # fetch from cache, if applicable
        if self.caches_target:
            return super().get_target_object(key)

        # okay now must fetch via API
        if self.get_keys() != ["uuid"]:
            raise ValueError("must use uuid key for this to work")
        uuid = key[0]

        try:
            result = self.farmos_client.resource.get_id(
                "taxonomy_term", "animal_type", str(uuid)
            )
        except requests.HTTPError as exc:
            if exc.response.status_code == 404:
                return None
            raise
        return result["data"]

    def normalize_target_object(self, obj):
        return {
            "uuid": UUID(obj["id"]),
            "name": obj["attributes"]["name"],
        }

    def get_type_payload(self, source_data):
        return {
            "attributes": {
                "name": source_data["name"],
            }
        }

    def create_target_object(self, key, source_data):
        if source_data.get("__ignoreme__"):
            return None
        if self.dry_run:
            return source_data

        payload = self.get_type_payload(source_data)
        result = self.farmos_client.resource.send(
            "taxonomy_term", "animal_type", payload
        )
        normal = self.normalize_target_object(result["data"])
        normal["_new_object"] = result["data"]
        return normal

    def update_target_object(self, asset, source_data, target_data=None):
        if self.dry_run:
            return asset

        payload = self.get_type_payload(source_data)
        payload["id"] = str(source_data["uuid"])
        result = self.farmos_client.resource.send(
            "taxonomy_term", "animal_type", payload
        )
        return self.normalize_target_object(result["data"])


class GroupAssetImporter(ToFarmOSAsset):

    model_title = "GroupAsset"
    farmos_asset_type = "group"

    supported_fields = [
        "uuid",
        "asset_name",
        "notes",
        "archived",
    ]


class LandAssetImporter(ToFarmOSAsset):

    model_title = "LandAsset"
    farmos_asset_type = "land"

    supported_fields = [
        "uuid",
        "asset_name",
        "land_type_id",
        "is_location",
        "is_fixed",
        "notes",
        "archived",
    ]

    def normalize_target_object(self, land):
        data = super().normalize_target_object(land)
        data.update(
            {
                "land_type_id": land["attributes"]["land_type"],
            }
        )
        return data

    def get_asset_payload(self, source_data):
        payload = super().get_asset_payload(source_data)

        attrs = {}
        if "land_type_id" in self.fields:
            attrs["land_type"] = source_data["land_type_id"]

        if attrs:
            payload["attributes"].update(attrs)

        return payload


class StructureAssetImporter(ToFarmOSAsset):

    model_title = "StructureAsset"
    farmos_asset_type = "structure"

    supported_fields = [
        "uuid",
        "asset_name",
        "structure_type_id",
        "is_location",
        "is_fixed",
        "notes",
        "archived",
    ]

    def normalize_target_object(self, structure):
        data = super().normalize_target_object(structure)
        data.update(
            {
                "structure_type_id": structure["attributes"]["structure_type"],
            }
        )
        return data

    def get_asset_payload(self, source_data):
        payload = super().get_asset_payload(source_data)

        attrs = {}
        if "structure_type_id" in self.fields:
            attrs["structure_type"] = source_data["structure_type_id"]

        if attrs:
            payload["attributes"].update(attrs)

        return payload



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
View for farmOS activity logs
"""

import datetime

import colander

from wuttaweb.forms.schema import WuttaDateTime
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views.farmos import FarmOSMasterView


class ActivityLogView(FarmOSMasterView):
    """
    View for farmOS activity logs
    """

    model_name = "farmos_activity_log"
    model_title = "farmOS Activity Log"
    model_title_plural = "farmOS Activity Logs"

    route_prefix = "farmos_logs_activity"
    url_prefix = "/farmOS/logs/activity"

    farmos_refurl_path = "/logs/activity"

    grid_columns = [
        "name",
        "timestamp",
        "status",
    ]

    sort_defaults = ("timestamp", "desc")

    form_fields = [
        "name",
        "timestamp",
        "status",
        "notes",
    ]

    def get_grid_data(self, columns=None, session=None):
        logs = self.farmos_client.log.get("activity")
        return [self.normalize_log(t) for t in logs["data"]]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")
        g.set_searchable("name")

        # timestamp
        g.set_renderer("timestamp", "datetime")

    def get_instance(self):
        log = self.farmos_client.log.get_id("activity", self.request.matchdict["uuid"])
        self.raw_json = log
        return self.normalize_log(log["data"])

    def get_instance_title(self, log):
        return log["name"]

    def normalize_log(self, log):

        if timestamp := log["attributes"]["timestamp"]:
            timestamp = datetime.datetime.fromisoformat(timestamp)
            timestamp = self.app.localtime(timestamp)

        if notes := log["attributes"]["notes"]:
            notes = notes["value"]

        return {
            "uuid": log["id"],
            "drupal_id": log["attributes"]["drupal_internal__id"],
            "name": log["attributes"]["name"],
            "timestamp": timestamp,
            "status": log["attributes"]["status"],
            "notes": notes or colander.null,
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # timestamp
        f.set_node("timestamp", WuttaDateTime())
        f.set_widget("timestamp", WuttaDateTimeWidget(self.request))

        # notes
        f.set_widget("notes", "notes")

    def get_xref_buttons(self, log):
        model = self.app.model
        session = self.Session()

        buttons = [
            self.make_button(
                "View in farmOS",
                primary=True,
                url=self.app.get_farmos_url(f"/log/{log['drupal_id']}"),
                target="_blank",
                icon_left="external-link-alt",
            ),
        ]

        if wf_log := (
            session.query(model.ActivityLog)
            .filter(model.ActivityLog.farmos_uuid == log["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url("activity_logs.view", uuid=wf_log.uuid),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    ActivityLogView = kwargs.get("ActivityLogView", base["ActivityLogView"])
    ActivityLogView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
farmOS integration handler
"""

from farmOS import farmOS

from wuttjamaican.app import GenericHandler


class FarmOSHandler(GenericHandler):
    """
    Base class and default implementation for the farmOS integration
    :term:`handler`.
    """

    def get_farmos_client(self, hostname=None, **kwargs):
        """
        Returns a new farmOS API client.
        """
        if not hostname:
            hostname = self.get_farmos_url()
        return farmOS(hostname, **kwargs)

    def get_farmos_version(self, client=None, *args, **kwargs):
        """
        Returns the farmOS version in use.
        """
        if not client:
            client = self.get_farmos_client(*args, **kwargs)
        info = client.info()
        return info["meta"]["farm"]["version"]

    def is_farmos_3x(self, client=None, *args, **kwargs):
        """
        Check if the farmOS version is 3.x.
        """
        if not client:
            client = self.get_farmos_client(*args, **kwargs)

        version = self.get_farmos_version(client)
        return version[0] == "3"

    def is_farmos_4x(self, client=None, *args, **kwargs):
        """
        Check if the farmOS version is 4.x.
        """
        if not client:
            client = self.get_farmos_client(*args, **kwargs)

        version = self.get_farmos_version(client)
        return version[0] == "4"

    def get_farmos_url(self, path=None, require=True):
        """
        Returns the base URL for farmOS, or one with ``path`` appended.

        Note that if no path is provided, the final slash will be
        stripped from the base URL.

        :param path: Optional path to append to the base URL.

        :param require: If true, an error is raised when base URL
           cannot be determined.

        :returns: URL string
        """
        base = self.config.get("farmos.url.base", require=require)
        if base:
            base = base.rstrip("/")

        if path:
            path = path.lstrip("/")
            return f"{base}/{path}"

        return base



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
View for farmOS animal types
"""

import datetime

import colander

from wuttaweb.forms.schema import WuttaDateTime

from wuttafarm.web.views.farmos import FarmOSMasterView


class AnimalTypeView(FarmOSMasterView):
    """
    Master view for Animal Types in farmOS.
    """

    model_name = "farmos_animal_type"
    model_title = "farmOS Animal Type"
    model_title_plural = "farmOS Animal Types"

    route_prefix = "farmos_animal_types"
    url_prefix = "/farmOS/animal-types"

    farmos_refurl_path = "/admin/structure/taxonomy/manage/animal_type/overview"

    grid_columns = [
        "name",
        "description",
        "changed",
    ]

    sort_defaults = "name"

    form_fields = [
        "name",
        "description",
        "changed",
    ]

    def get_grid_data(self, columns=None, session=None):
        animal_types = self.farmos_client.resource.get("taxonomy_term", "animal_type")
        return [self.normalize_animal_type(t) for t in animal_types["data"]]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")
        g.set_searchable("name")

        # changed
        g.set_renderer("changed", "datetime")

    def get_instance(self):
        animal_type = self.farmos_client.resource.get_id(
            "taxonomy_term", "animal_type", self.request.matchdict["uuid"]
        )
        self.raw_json = animal_type
        return self.normalize_animal_type(animal_type["data"])

    def get_instance_title(self, animal_type):
        return animal_type["name"]

    def normalize_animal_type(self, animal_type):

        if changed := animal_type["attributes"]["changed"]:
            changed = datetime.datetime.fromisoformat(changed)
            changed = self.app.localtime(changed)

        if description := animal_type["attributes"]["description"]:
            description = description["value"]

        return {
            "uuid": animal_type["id"],
            "drupal_id": animal_type["attributes"]["drupal_internal__tid"],
            "name": animal_type["attributes"]["name"],
            "description": description or colander.null,
            "changed": changed,
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # description
        f.set_widget("description", "notes")

        # changed
        f.set_node("changed", WuttaDateTime())

    def get_xref_buttons(self, animal_type):
        model = self.app.model
        session = self.Session()

        buttons = [
            self.make_button(
                "View in farmOS",
                primary=True,
                url=self.app.get_farmos_url(
                    f"/taxonomy/term/{animal_type['drupal_id']}"
                ),
                target="_blank",
                icon_left="external-link-alt",
            )
        ]

        if wf_animal_type := (
            session.query(model.AnimalType)
            .filter(model.AnimalType.farmos_uuid == animal_type["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "animal_types.view", uuid=wf_animal_type.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    AnimalTypeView = kwargs.get("AnimalTypeView", base["AnimalTypeView"])
    AnimalTypeView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Model definition for Structure Types
"""

import sqlalchemy as sa
from sqlalchemy import orm

from wuttjamaican.db import model

from wuttafarm.db.model.assets import AssetMixin, add_asset_proxies


class StructureType(model.Base):
    """
    Represents a "structure type" from farmOS
    """

    __tablename__ = "structure_type"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Structure Type",
        "model_title_plural": "Structure Types",
    }

    uuid = model.uuid_column()

    name = sa.Column(
        sa.String(length=100),
        nullable=False,
        unique=True,
        doc="""
        Name of the structure type.
        """,
    )

    farmos_uuid = sa.Column(
        model.UUID(),
        nullable=True,
        unique=True,
        doc="""
        UUID for the structure type within farmOS.
        """,
    )

    drupal_id = sa.Column(
        sa.String(length=50),
        nullable=True,
        unique=True,
        doc="""
        Drupal internal ID for the structure type.
        """,
    )

    def __str__(self):
        return self.name or ""


class StructureAsset(AssetMixin, model.Base):
    """
    Represents a structure from farmOS
    """

    __tablename__ = "asset_structure"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Structure Asset",
        "model_title_plural": "Structure Assets",
        "farmos_asset_type": "structure",
    }

    structure_type_uuid = model.uuid_fk_column("structure_type.uuid", nullable=False)
    structure_type = orm.relationship(
        "StructureType",
        doc="""
        Reference to the type of structure.
        """,
    )


add_asset_proxies(StructureAsset)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Custom app handler for WuttaFarm
"""

from wuttjamaican import app as base


class WuttaFarmAppHandler(base.AppHandler):
    """
    Custom :term:`app handler` for WuttaFarm.
    """

    display_format_datetime = "%a, %m/%d/%Y - %H:%M"

    default_auth_handler_spec = "wuttafarm.auth:WuttaFarmAuthHandler"
    default_install_handler_spec = "wuttafarm.install:WuttaFarmInstallHandler"

    def get_farmos_handler(self):
        """
        Get the configured farmOS integration handler.

        :rtype: :class:`~wuttafarm.farmos.FarmOSHandler`
        """
        if "farmos" not in self.handlers:
            spec = self.config.get(
                f"{self.appname}.farmos_handler",
                default="wuttafarm.farmos.handler:FarmOSHandler",
            )
            factory = self.load_object(spec)
            self.handlers["farmos"] = factory(self.config)
        return self.handlers["farmos"]

    def get_farmos_url(self, *args, **kwargs):
        """
        Get a farmOS URL.  This is a convenience wrapper around
        :meth:`~wuttafarm.farmos.handler.FarmOSHandler.get_farmos_url()`.
        """
        handler = self.get_farmos_handler()
        return handler.get_farmos_url(*args, **kwargs)

    def get_farmos_client(self, *args, **kwargs):
        """
        Get a farmOS client.  This is a convenience wrapper around
        :meth:`~wuttafarm.farmos.handler.FarmOSHandler.get_farmos_client()`.
        """
        handler = self.get_farmos_handler()
        return handler.get_farmos_client(*args, **kwargs)

    def is_farmos_3x(self, *args, **kwargs):
        """
        Check if the farmOS version is 3.x.  This is a convenience
        wrapper around
        :meth:`~wuttafarm.farmos.handler.FarmOSHandler.is_farmos_3x()`.
        """
        handler = self.get_farmos_handler()
        return handler.is_farmos_3x(*args, **kwargs)

    def is_farmos_4x(self, *args, **kwargs):
        """
        Check if the farmOS version is 4.x.  This is a convenience
        wrapper around
        :meth:`~wuttafarm.farmos.handler.FarmOSHandler.is_farmos_4x()`.
        """
        handler = self.get_farmos_handler()
        return handler.is_farmos_4x(*args, **kwargs)

    def export_to_farmos(self, obj, require=True):
        """
        Export the given object to farmOS, using configured handler.

        This should ensure the given object is also *updated* with the
        farmOS UUID and Drupal ID, when new record is created in
        farmOS.

        :param obj: Any data object in WuttaFarm, e.g. AnimalAsset
           instance.

        :param require: If true, this will *require* the export
           handler to support objects of the given type.  If false,
           then nothing will happen / export is silently skipped when
           there is no such exporter.
        """
        handler = self.app.get_import_handler("export.to_farmos.from_wuttafarm")

        model_name = type(obj).__name__
        if model_name not in handler.importers:
            if require:
                raise ValueError(f"no exporter found for {model_name}")
            return

        # nb. begin txn to establish the API client
        # TODO: should probably use current user oauth2 token instead
        # of always making a new one here, which is what happens IIUC
        handler.begin_target_transaction()
        importer = handler.get_importer(model_name, caches_target=False)
        normal = importer.normalize_source_object(obj)
        importer.process_data(source_data=[normal])


class WuttaFarmAppProvider(base.AppProvider):
    """
    The :term:`app provider` for WuttaFarm.
    """

    email_modules = ["wuttafarm.emails"]



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Custom form widgets for WuttaFarm
"""

import json

import colander
from deform.widget import Widget
from webhelpers2.html import HTML, tags

from wuttaweb.forms.widgets import WuttaCheckboxChoiceWidget
from wuttaweb.db import Session


class ImageWidget(Widget):
    """
    Widget to display an image URL for a record.
    """

    def __init__(self, alt_text, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alt_text = alt_text

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            return tags.image(cstruct, self.alt_text, **kw)

        return super().serialize(field, cstruct, **kw)


class AnimalTypeWidget(Widget):
    """
    Widget to display an "animal type" field.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            animal_type = json.loads(cstruct)
            return tags.link_to(
                animal_type["name"],
                self.request.route_url(
                    "farmos_animal_types.view", uuid=animal_type["uuid"]
                ),
            )

        return super().serialize(field, cstruct, **kw)


class StructureWidget(Widget):
    """
    Widget to display a "structure" field.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            structure = json.loads(cstruct)
            return tags.link_to(
                structure["name"],
                self.request.route_url(
                    "farmos_structures.view", uuid=structure["uuid"]
                ),
            )

        return super().serialize(field, cstruct, **kw)


class UsersWidget(Widget):
    """
    Widget to display the list of owners for an asset etc.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            items = []
            for user in json.loads(cstruct):
                link = tags.link_to(
                    user["display_name"],
                    self.request.route_url("farmos_users.view", uuid=user["uuid"]),
                )
                items.append(HTML.tag("li", c=link))

            return HTML.tag("ul", c=items)

        return super().serialize(field, cstruct, **kw)


class AssetParentRefsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for Parents field which references assets.
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        model = self.app.model
        session = Session()

        readonly = kw.get("readonly", self.readonly)
        if readonly:
            parents = []
            for uuid in json.loads(cstruct):
                parent = session.get(model.Asset, uuid)
                parents.append(
                    HTML.tag(
                        "li",
                        c=tags.link_to(
                            str(parent),
                            self.request.route_url(
                                f"{parent.asset_type}_assets.view", uuid=parent.uuid
                            ),
                        ),
                    )
                )
            return HTML.tag("ul", c=parents)

        return super().serialize(field, cstruct, **kw)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Auth views
"""

from oauthlib.oauth2 import AccessDeniedError
from requests_oauthlib import OAuth2Session

from wuttaweb.views import auth as base
from wuttaweb.auth import login_user
from wuttaweb.db import Session

from wuttafarm.web.util import save_farmos_oauth2_token


class AuthView(base.AuthView):
    """
    Auth views
    """

    def authenticate_user(self, session, username, password):
        if user := super().authenticate_user(session, username, password):

            # nb. auth handler will stash the farmOS oauth2 token in
            # the user object, if applicable (i.e. unless the user
            # authenticated normally via the app DB).  when a token is
            # involved we need to put that in the user's web session.
            if "farmos_oauth2_token" in user.__dict__:
                self.request.session["farmos.oauth2.token"] = user.__dict__.pop(
                    "farmos_oauth2_token"
                )

            return user

        return None

    def get_farmos_oauth2_session(self):
        return OAuth2Session(
            client_id="farm",
            scope="farm_manager",
            redirect_uri=self.request.route_url("farmos_oauth_callback"),
        )

    def farmos_oauth_login(self):
        """
        View to initiate OAuth2 workflow.
        """
        oauth = self.get_farmos_oauth2_session()
        auth_url, state = oauth.authorization_url(
            self.app.get_farmos_url("/oauth/authorize")
        )
        return self.redirect(auth_url)

    def farmos_oauth_callback(self):
        """
        View for OAuth2 workflow, provided as redirect URL when
        authorizing.
        """
        session = Session()
        auth = self.app.get_auth_handler()
        oauth = self.get_farmos_oauth2_session()

        try:
            # get oauth token from farmOS
            token = oauth.fetch_token(
                self.app.get_farmos_url("/oauth/token"),
                authorization_response=self.request.current_route_url(),
                include_client_id=True,
            )

        except AccessDeniedError:
            self.request.session.flash("Access to farmOS was denied.", "error")
            return self.redirect(self.request.route_url("login"))

        # save token in user session
        save_farmos_oauth2_token(self.request, token)

        # nb. must give a *copy* of the token to farmOS client, since
        # it will mutate it in-place and we don't want that to happen
        # for our original copy in the user session.  (otherwise the
        # auto-refresh will not work correctly for subsequent calls.)
        token = dict(token)

        # get (or create) native app user
        farmos_client = self.app.get_farmos_client(token=token)
        info = farmos_client.info()
        farmos_user = farmos_client.resource.get_id(
            "user", "user", info["meta"]["links"]["me"]["meta"]["id"]
        )
        user = auth.get_or_make_farmos_user(
            session, farmos_user["data"]["attributes"]["name"]
        )
        if not user:
            self.request.session.flash(
                "farmOS authentication was successful, but user is "
                "not allowed login to {self.app.get_node_title()}.",
                "error",
            )
            return self.redirect(self.request.route_url("login"))

        # delare user is logged in
        headers = login_user(self.request, user)
        referrer = self.request.get_referrer()
        return self.redirect(referrer, headers=headers)

    @classmethod
    def defaults(cls, config):
        cls._auth_defaults(config)
        cls._wuttafarm_defaults(config)

    @classmethod
    def _wuttafarm_defaults(cls, config):

        # farmos oauth login
        config.add_route("farmos_oauth_login", "/farmos/oauth/login")
        config.add_view(cls, attr="farmos_oauth_login", route_name="farmos_oauth_login")

        # farmos oauth callback
        config.add_route("farmos_oauth_callback", "/farmos/oauth/callback")
        config.add_view(
            cls, attr="farmos_oauth_callback", route_name="farmos_oauth_callback"
        )


def defaults(config, **kwargs):
    local = globals()
    AuthView = kwargs.get("AuthView", local["AuthView"])
    base.defaults(config, **{"AuthView": AuthView})


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base class for farmOS master views
"""

import json

import markdown

from wuttaweb.views import MasterView

from wuttafarm.web.util import save_farmos_oauth2_token


class FarmOSMasterView(MasterView):
    """
    Base class for farmOS master views
    """

    model_key = "uuid"

    creatable = False
    editable = False
    deletable = False

    filterable = False
    sort_on_backend = False
    paginate_on_backend = False

    farmos_refurl_path = None

    labels = {
        "raw_image_url": "Raw Image URL",
        "large_image_url": "Large Image URL",
        "thumbnail_image_url": "Thumbnail Image URL",
    }

    def __init__(self, request, context=None):
        super().__init__(request, context=context)
        self.farmos_client = self.get_farmos_client()
        self.farmos_4x = self.app.is_farmos_4x(self.farmos_client)
        self.raw_json = None

    def get_farmos_client(self):
        token = self.request.session.get("farmos.oauth2.token")
        if not token:
            raise self.forbidden()

        # nb. must give a *copy* of the token to farmOS client, since
        # it will mutate it in-place and we don't want that to happen
        # for our original copy in the user session.  (otherwise the
        # auto-refresh will not work correctly for subsequent calls.)
        token = dict(token)

        def token_updater(token):
            save_farmos_oauth2_token(self.request, token)

        return self.app.get_farmos_client(token=token, token_updater=token_updater)

    def get_fallback_templates(self, template):
        """ """
        templates = super().get_fallback_templates(template)

        if template == "view":
            templates.insert(0, "/farmos/master/view.mako")

        return templates

    def get_template_context(self, context):

        if self.listing and self.farmos_refurl_path:
            context["farmos_refurl"] = self.app.get_farmos_url(self.farmos_refurl_path)

        if self.viewing and self.raw_json:
            context["raw_json"] = self.raw_json
            code = "```json\n" + json.dumps(self.raw_json, indent=2) + "\n```"
            # TODO: this does not seem to be adding syntax highlight
            context["rendered_json"] = markdown.markdown(
                code, extensions=["fenced_code", "codehilite"]
            )

        return context



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaFarm → farmOS data export
"""

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

from wuttasync.importing import ImportHandler, FromWuttaHandler, FromWutta, Orientation

from wuttafarm.db import model
from wuttafarm.farmos import importing as farmos_importing


class FromWuttaFarmHandler(FromWuttaHandler):
    """
    Base class for import handler targeting WuttaFarm
    """

    source_key = "wuttafarm"


class ToFarmOSHandler(ImportHandler):
    """
    Base class for export handlers using CSV file(s) as data target.
    """

    target_key = "farmos"
    generic_target_title = "farmOS"

    # TODO: a lot of duplication to cleanup here; see FromFarmOSHandler

    def begin_target_transaction(self):
        """
        Establish the farmOS API client.
        """
        token = self.get_farmos_oauth2_token()
        self.farmos_client = self.app.get_farmos_client(token=token)
        self.farmos_4x = self.app.is_farmos_4x(self.farmos_client)

    def get_farmos_oauth2_token(self):

        client_id = self.config.get(
            "farmos.oauth2.importing.client_id", default="wuttafarm"
        )
        client_secret = self.config.require("farmos.oauth2.importing.client_secret")
        scope = self.config.get("farmos.oauth2.importing.scope", default="farm_manager")

        client = BackendApplicationClient(client_id=client_id)
        oauth = OAuth2Session(client=client)

        return oauth.fetch_token(
            token_url=self.app.get_farmos_url("/oauth/token"),
            include_client_id=True,
            client_secret=client_secret,
            scope=scope,
        )

    def get_importer_kwargs(self, key, **kwargs):
        kwargs = super().get_importer_kwargs(key, **kwargs)
        kwargs["farmos_client"] = self.farmos_client
        kwargs["farmos_4x"] = self.farmos_4x
        return kwargs


class FromWuttaFarmToFarmOS(FromWuttaFarmHandler, ToFarmOSHandler):
    """
    Handler for WuttaFarm → farmOS API export.
    """

    orientation = Orientation.EXPORT

    def define_importers(self):
        """ """
        importers = super().define_importers()
        importers["LandAsset"] = LandAssetImporter
        importers["StructureAsset"] = StructureAssetImporter
        importers["AnimalType"] = AnimalTypeImporter
        importers["AnimalAsset"] = AnimalAssetImporter
        importers["GroupAsset"] = GroupAssetImporter
        return importers


class FromWuttaFarm(FromWutta):

    drupal_internal_id_field = "drupal_internal__id"

    def create_target_object(self, key, source_data):
        obj = super().create_target_object(key, source_data)
        if obj is None:
            return None

        if not self.dry_run:

            # set farmOS, Drupal key fields in WuttaFarm
            api_object = obj["_new_object"]
            wf_object = source_data["_src_object"]
            wf_object.farmos_uuid = obj["uuid"]
            wf_object.drupal_id = api_object["attributes"][
                self.drupal_internal_id_field
            ]

        return obj


class AnimalAssetImporter(FromWuttaFarm, farmos_importing.model.AnimalAssetImporter):
    """
    WuttaFarm → farmOS API exporter for Animal Assets
    """

    source_model_class = model.AnimalAsset

    supported_fields = [
        "uuid",
        "asset_name",
        "animal_type_uuid",
        "sex",
        "is_sterile",
        "birthdate",
        "notes",
        "archived",
    ]

    def normalize_source_object(self, animal):
        return {
            "uuid": animal.farmos_uuid or self.app.make_true_uuid(),
            "asset_name": animal.asset_name,
            "animal_type_uuid": animal.animal_type.farmos_uuid,
            "sex": animal.sex,
            "is_sterile": animal.is_sterile,
            "birthdate": animal.birthdate,
            "notes": animal.notes,
            "archived": animal.archived,
            "_src_object": animal,
        }


class AnimalTypeImporter(FromWuttaFarm, farmos_importing.model.AnimalTypeImporter):
    """
    WuttaFarm → farmOS API exporter for Animal Types
    """

    source_model_class = model.AnimalType

    supported_fields = [
        "uuid",
        "name",
    ]

    drupal_internal_id_field = "drupal_internal__tid"

    def normalize_source_object(self, animal_type):
        return {
            "uuid": animal_type.farmos_uuid or self.app.make_true_uuid(),
            "name": animal_type.name,
            "_src_object": animal_type,
        }


class GroupAssetImporter(FromWuttaFarm, farmos_importing.model.GroupAssetImporter):
    """
    WuttaFarm → farmOS API exporter for Group Assets
    """

    source_model_class = model.GroupAsset

    supported_fields = [
        "uuid",
        "asset_name",
        "notes",
        "archived",
    ]

    def normalize_source_object(self, group):
        return {
            "uuid": group.farmos_uuid or self.app.make_true_uuid(),
            "asset_name": group.asset_name,
            "notes": group.notes,
            "archived": group.archived,
            "_src_object": group,
        }


class LandAssetImporter(FromWuttaFarm, farmos_importing.model.LandAssetImporter):
    """
    WuttaFarm → farmOS API exporter for Land Assets
    """

    source_model_class = model.LandAsset

    supported_fields = [
        "uuid",
        "asset_name",
        "land_type_id",
        "is_location",
        "is_fixed",
        "notes",
        "archived",
    ]

    def normalize_source_object(self, land):
        return {
            "uuid": land.farmos_uuid or self.app.make_true_uuid(),
            "asset_name": land.asset_name,
            "land_type_id": land.land_type.drupal_id,
            "is_location": land.is_location,
            "is_fixed": land.is_fixed,
            "notes": land.notes,
            "archived": land.archived,
            "_src_object": land,
        }


class StructureAssetImporter(
    FromWuttaFarm, farmos_importing.model.StructureAssetImporter
):
    """
    WuttaFarm → farmOS API exporter for Structure Assets
    """

    source_model_class = model.StructureAsset

    supported_fields = [
        "uuid",
        "asset_name",
        "structure_type_id",
        "is_location",
        "is_fixed",
        "notes",
        "archived",
    ]

    def normalize_source_object(self, structure):
        return {
            "uuid": structure.farmos_uuid or self.app.make_true_uuid(),
            "asset_name": structure.asset_name,
            "structure_type_id": structure.structure_type.drupal_id,
            "is_location": structure.is_location,
            "is_fixed": structure.is_fixed,
            "notes": structure.notes,
            "archived": structure.archived,
            "_src_object": structure,
        }



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Auth handler for use with farmOS
"""

from uuid import UUID

from oauthlib.oauth2.rfc6749.errors import InvalidGrantError
from sqlalchemy import orm

from wuttjamaican.auth import AuthHandler


class WuttaFarmAuthHandler(AuthHandler):
    """
    Custom auth handler for WuttaFarm.  This adds some magic around
    the user login process.

    It can attempt authentication against the configured farmOS
    instance, auto-creating native users within the app DB as needed.

    """

    def get_role_farm_manager(self, session):
        """
        Returns the special "Farm Manager" role.
        """
        return self._special_role(
            session, UUID("06979646-b1b9-723b-8000-b86161427f9f"), "Farm Manager"
        )

    def get_role_farm_worker(self, session):
        """
        Returns the special "Farm Worker" role.
        """
        return self._special_role(
            session, UUID("06979648-9bd6-7b61-8000-ee8818052ee8"), "Farm Worker"
        )

    def get_role_farm_viewer(self, session):
        """
        Returns the special "Farm Viewer" role.
        """
        return self._special_role(
            session, UUID("06979649-ed32-7e97-8000-113ddf0ab5f3"), "Farm Viewer"
        )

    def authenticate_user(self, session, username, password):
        """
        When authentication is attempted, this first will check
        credentials against the app DB per normal logic.  If that
        succeeds, the result is no different from the typical
        behavior.

        If default logic fails, this will try to obtain an OAuth2
        token from the farmOS site.  If that succeeds, then a lookup
        is done in the app DB for a matching user.  If the user does
        not yet exist it will be created automatically.
        """
        if user := super().authenticate_user(session, username, password):
            return user

        if token := self.get_farmos_oauth2_token(username, password):
            if user := self.get_or_make_farmos_user(session, username):
                user.__dict__["farmos_oauth2_token"] = token
                return user

        return None

    def get_farmos_oauth2_token(self, username, password):
        client = self.app.get_farmos_client()
        try:
            return client.authorize(username=username, password=password)
        except InvalidGrantError:
            return None

    def get_or_make_farmos_user(self, session, username):
        model = self.app.model
        try:
            user = (
                session.query(model.User).filter(model.User.username == username).one()
            )
        except orm.exc.NoResultFound:
            pass
        else:
            return user if user.active else None

        # nb. prevent edit for farmOS mirrored user accounts
        user = self.make_user(session, username=username, prevent_edit=True)

        # TODO
        manager = self.get_role_farm_manager(session)
        user.roles.append(manager)

        return user



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaFarm web app
"""

import os

from wuttaweb import app as base


def main(global_config, **settings):
    """
    Make and return the WSGI app (Paste entry point).
    """
    # prefer WuttaFarm templates over wuttaweb
    settings.setdefault(
        "mako.directories",
        [
            "wuttafarm.web:templates",
            "wuttaweb:templates",
        ],
    )

    # make config objects
    wutta_config = base.make_wutta_config(settings)
    pyramid_config = base.make_pyramid_config(settings)
    app = wutta_config.get_app()

    # custom buefy css
    # TODO: this is not robust enough, probably..but works for me/now
    wutta_config.setdefault(
        "wuttaweb.liburl.buefy_css", "/wuttafarm/css/wuttafarm-buefy.css"
    )
    # nb. ensure buefy version matches what we use for custom css
    wutta_config.setdefault("wuttaweb.libver.buefy", "0.9.29")
    wutta_config.setdefault("wuttaweb.libver.buefy_css", "0.9.29")

    # bring in the rest of WuttaFarm
    pyramid_config.include("wuttafarm.web.static")
    pyramid_config.include("wuttafarm.web.subscribers")
    pyramid_config.include("wuttafarm.web.views")

    return pyramid_config.make_wsgi_app()


def make_wsgi_app():
    """
    Make and return the WSGI app (generic entry point).
    """
    return base.make_wsgi_app(main)


def make_asgi_app():
    """
    Make and return the ASGI app (generic entry point).
    """
    return base.make_asgi_app(main)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Activity Logs
"""

from wuttafarm.db.model.logs import ActivityLog
from wuttafarm.web.views import WuttaFarmMasterView


class ActivityLogView(WuttaFarmMasterView):
    """
    Master view for Activity Logs
    """

    model_class = ActivityLog
    route_prefix = "activity_logs"
    url_prefix = "/logs/activity"

    farmos_refurl_path = "/logs/activity"

    grid_columns = [
        "message",
        "timestamp",
        "status",
    ]

    sort_defaults = ("timestamp", "desc")

    filter_defaults = {
        "message": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "message",
        "timestamp",
        "status",
        "notes",
        "farmos_uuid",
        "drupal_id",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # message
        g.set_link("message")

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # notes
        f.set_widget("notes", "notes")

    def get_farmos_url(self, log):
        return self.app.get_farmos_url(f"/log/{log.drupal_id}")

    def get_xref_buttons(self, log):
        buttons = super().get_xref_buttons(log)

        if log.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_logs_activity.view", uuid=log.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    ActivityLogView = kwargs.get("ActivityLogView", base["ActivityLogView"])
    ActivityLogView.defaults(config)


def includeme(config):
    defaults(config)



"""use shared base for Land Assets

Revision ID: d882682c82f9
Revises: d6e8d16d6854
Create Date: 2026-02-15 12:00:27.036011

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "d882682c82f9"
down_revision: Union[str, None] = "d6e8d16d6854"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # asset_parent
    op.create_table(
        "asset_parent",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("asset_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("parent_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["asset_uuid"],
            ["asset.uuid"],
            name=op.f("fk_asset_parent_asset_uuid_asset"),
        ),
        sa.ForeignKeyConstraint(
            ["parent_uuid"],
            ["asset.uuid"],
            name=op.f("fk_asset_parent_parent_uuid_asset"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_asset_parent")),
    )
    op.create_table(
        "asset_parent_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "asset_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "parent_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_asset_parent_version")
        ),
    )
    op.create_index(
        op.f("ix_asset_parent_version_end_transaction_id"),
        "asset_parent_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_parent_version_operation_type"),
        "asset_parent_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_asset_parent_version_pk_transaction_id",
        "asset_parent_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_asset_parent_version_pk_validity",
        "asset_parent_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_parent_version_transaction_id"),
        "asset_parent_version",
        ["transaction_id"],
        unique=False,
    )

    # asset_land
    op.create_table(
        "asset_land",
        sa.Column("land_type_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["land_type_uuid"],
            ["land_type.uuid"],
            name=op.f("fk_asset_land_land_type_uuid_land_type"),
        ),
        sa.ForeignKeyConstraint(
            ["uuid"], ["asset.uuid"], name=op.f("fk_asset_land_uuid_asset")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_asset_land")),
        sa.UniqueConstraint(
            "land_type_uuid", name=op.f("uq_asset_land_land_type_uuid")
        ),
    )
    op.create_table(
        "asset_land_version",
        sa.Column(
            "land_type_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_asset_land_version")
        ),
    )
    op.create_index(
        op.f("ix_asset_land_version_end_transaction_id"),
        "asset_land_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_land_version_operation_type"),
        "asset_land_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_asset_land_version_pk_transaction_id",
        "asset_land_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_asset_land_version_pk_validity",
        "asset_land_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_land_version_transaction_id"),
        "asset_land_version",
        ["transaction_id"],
        unique=False,
    )

    # land_asset_parent
    op.drop_index(
        op.f("ix_land_asset_parent_version_end_transaction_id"),
        table_name="land_asset_parent_version",
    )
    op.drop_index(
        op.f("ix_land_asset_parent_version_operation_type"),
        table_name="land_asset_parent_version",
    )
    op.drop_index(
        op.f("ix_land_asset_parent_version_pk_transaction_id"),
        table_name="land_asset_parent_version",
    )
    op.drop_index(
        op.f("ix_land_asset_parent_version_pk_validity"),
        table_name="land_asset_parent_version",
    )
    op.drop_index(
        op.f("ix_land_asset_parent_version_transaction_id"),
        table_name="land_asset_parent_version",
    )
    op.drop_table("land_asset_parent_version")
    op.drop_table("land_asset_parent")

    # land_asset
    op.drop_index(
        op.f("ix_land_asset_version_end_transaction_id"),
        table_name="land_asset_version",
    )
    op.drop_index(
        op.f("ix_land_asset_version_operation_type"), table_name="land_asset_version"
    )
    op.drop_index(
        op.f("ix_land_asset_version_pk_transaction_id"), table_name="land_asset_version"
    )
    op.drop_index(
        op.f("ix_land_asset_version_pk_validity"), table_name="land_asset_version"
    )
    op.drop_index(
        op.f("ix_land_asset_version_transaction_id"), table_name="land_asset_version"
    )
    op.drop_table("land_asset_version")
    op.drop_table("land_asset")


def downgrade() -> None:

    # land_asset
    op.create_table(
        "land_asset",
        sa.Column("uuid", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column("name", sa.VARCHAR(length=100), autoincrement=False, nullable=False),
        sa.Column("land_type_uuid", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column("is_location", sa.BOOLEAN(), autoincrement=False, nullable=False),
        sa.Column("is_fixed", sa.BOOLEAN(), autoincrement=False, nullable=False),
        sa.Column("notes", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column("archived", sa.BOOLEAN(), autoincrement=False, nullable=False),
        sa.Column("farmos_uuid", sa.UUID(), autoincrement=False, nullable=True),
        sa.Column("drupal_id", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.ForeignKeyConstraint(
            ["land_type_uuid"],
            ["land_type.uuid"],
            name=op.f("fk_land_asset_land_type_uuid_land_type"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_land_asset")),
        sa.UniqueConstraint(
            "drupal_id",
            name=op.f("uq_land_asset_drupal_id"),
            postgresql_include=[],
            postgresql_nulls_not_distinct=False,
        ),
        sa.UniqueConstraint(
            "farmos_uuid",
            name=op.f("uq_land_asset_farmos_uuid"),
            postgresql_include=[],
            postgresql_nulls_not_distinct=False,
        ),
        sa.UniqueConstraint(
            "land_type_uuid",
            name=op.f("uq_land_asset_land_type_uuid"),
            postgresql_include=[],
            postgresql_nulls_not_distinct=False,
        ),
        sa.UniqueConstraint(
            "name",
            name=op.f("uq_land_asset_name"),
            postgresql_include=[],
            postgresql_nulls_not_distinct=False,
        ),
    )
    op.create_table(
        "land_asset_version",
        sa.Column("uuid", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column("name", sa.VARCHAR(length=100), autoincrement=False, nullable=True),
        sa.Column("land_type_uuid", sa.UUID(), autoincrement=False, nullable=True),
        sa.Column("is_location", sa.BOOLEAN(), autoincrement=False, nullable=True),
        sa.Column("is_fixed", sa.BOOLEAN(), autoincrement=False, nullable=True),
        sa.Column("notes", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column("archived", sa.BOOLEAN(), autoincrement=False, nullable=True),
        sa.Column("farmos_uuid", sa.UUID(), autoincrement=False, nullable=True),
        sa.Column("drupal_id", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column("transaction_id", sa.BIGINT(), autoincrement=False, nullable=False),
        sa.Column(
            "end_transaction_id", sa.BIGINT(), autoincrement=False, nullable=True
        ),
        sa.Column("operation_type", sa.SMALLINT(), autoincrement=False, nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_land_asset_version")
        ),
    )
    op.create_index(
        op.f("ix_land_asset_version_transaction_id"),
        "land_asset_version",
        ["transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_land_asset_version_pk_validity"),
        "land_asset_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_land_asset_version_pk_transaction_id"),
        "land_asset_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        op.f("ix_land_asset_version_operation_type"),
        "land_asset_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_land_asset_version_end_transaction_id"),
        "land_asset_version",
        ["end_transaction_id"],
        unique=False,
    )

    # land_asset_parent
    op.create_table(
        "land_asset_parent",
        sa.Column("uuid", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column("land_asset_uuid", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column("parent_asset_uuid", sa.UUID(), autoincrement=False, nullable=False),
        sa.ForeignKeyConstraint(
            ["land_asset_uuid"],
            ["land_asset.uuid"],
            name=op.f("fk_land_asset_parent_land_asset_uuid_land_asset"),
        ),
        sa.ForeignKeyConstraint(
            ["parent_asset_uuid"],
            ["land_asset.uuid"],
            name=op.f("fk_land_asset_parent_parent_asset_uuid_land_asset"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_land_asset_parent")),
    )
    op.create_table(
        "land_asset_parent_version",
        sa.Column("uuid", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column("land_asset_uuid", sa.UUID(), autoincrement=False, nullable=True),
        sa.Column("parent_asset_uuid", sa.UUID(), autoincrement=False, nullable=True),
        sa.Column("transaction_id", sa.BIGINT(), autoincrement=False, nullable=False),
        sa.Column(
            "end_transaction_id", sa.BIGINT(), autoincrement=False, nullable=True
        ),
        sa.Column("operation_type", sa.SMALLINT(), autoincrement=False, nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_land_asset_parent_version")
        ),
    )
    op.create_index(
        op.f("ix_land_asset_parent_version_transaction_id"),
        "land_asset_parent_version",
        ["transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_land_asset_parent_version_pk_validity"),
        "land_asset_parent_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_land_asset_parent_version_pk_transaction_id"),
        "land_asset_parent_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        op.f("ix_land_asset_parent_version_operation_type"),
        "land_asset_parent_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_land_asset_parent_version_end_transaction_id"),
        "land_asset_parent_version",
        ["end_transaction_id"],
        unique=False,
    )

    # asset_land
    op.drop_table("asset_land")
    op.drop_index(
        op.f("ix_asset_land_version_transaction_id"), table_name="asset_land_version"
    )
    op.drop_index("ix_asset_land_version_pk_validity", table_name="asset_land_version")
    op.drop_index(
        "ix_asset_land_version_pk_transaction_id", table_name="asset_land_version"
    )
    op.drop_index(
        op.f("ix_asset_land_version_operation_type"), table_name="asset_land_version"
    )
    op.drop_index(
        op.f("ix_asset_land_version_end_transaction_id"),
        table_name="asset_land_version",
    )
    op.drop_table("asset_land_version")

    # asset_parent
    op.drop_index(
        op.f("ix_asset_parent_version_transaction_id"),
        table_name="asset_parent_version",
    )
    op.drop_index(
        "ix_asset_parent_version_pk_validity", table_name="asset_parent_version"
    )
    op.drop_index(
        "ix_asset_parent_version_pk_transaction_id", table_name="asset_parent_version"
    )
    op.drop_index(
        op.f("ix_asset_parent_version_operation_type"),
        table_name="asset_parent_version",
    )
    op.drop_index(
        op.f("ix_asset_parent_version_end_transaction_id"),
        table_name="asset_parent_version",
    )
    op.drop_table("asset_parent_version")
    op.drop_table("asset_parent")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Groups
"""

from wuttafarm.web.views.assets import AssetMasterView
from wuttafarm.db.model.groups import GroupAsset


class GroupView(AssetMasterView):
    """
    Master view for Groups
    """

    model_class = GroupAsset
    route_prefix = "group_assets"
    url_prefix = "/assets/group"

    farmos_refurl_path = "/assets/group"

    grid_columns = [
        "thumbnail",
        "drupal_id",
        "asset_name",
        "archived",
    ]

    form_fields = [
        "asset_name",
        "notes",
        "asset_type",
        "archived",
        "farmos_uuid",
        "drupal_id",
    ]


def defaults(config, **kwargs):
    base = globals()

    GroupView = kwargs.get("GroupView", base["GroupView"])
    GroupView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Assets
"""

from collections import OrderedDict

from wuttaweb.forms.schema import WuttaDictEnum
from wuttaweb.db import Session

from wuttafarm.web.views import WuttaFarmMasterView
from wuttafarm.db.model import Asset
from wuttafarm.web.forms.schema import AssetParentRefs
from wuttafarm.web.forms.widgets import ImageWidget


def get_asset_type_enum(config):
    app = config.get_app()
    model = app.model
    session = Session()
    asset_types = OrderedDict()
    query = session.query(model.AssetType).order_by(model.AssetType.name)
    for asset_type in query:
        asset_types[asset_type.drupal_id] = asset_type.name
    return asset_types


class AssetView(WuttaFarmMasterView):
    """
    Master view for Assets
    """

    model_class = Asset
    route_prefix = "assets"
    url_prefix = "/assets"

    farmos_refurl_path = "/assets"

    viewable = False
    creatable = False
    editable = False
    deletable = False
    model_is_versioned = False

    grid_columns = [
        "thumbnail",
        "drupal_id",
        "asset_name",
        "asset_type",
        "parents",
        "archived",
    ]

    sort_defaults = "asset_name"

    filter_defaults = {
        "asset_name": {"active": True, "verb": "contains"},
        "archived": {"active": True, "verb": "is_false"},
    }

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # thumbnail
        g.set_renderer("thumbnail", self.render_grid_thumbnail)
        g.set_label("thumbnail", "", column_only=True)
        g.set_centered("thumbnail")

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)

        # asset_name
        g.set_link("asset_name")

        # asset_type
        g.set_enum("asset_type", get_asset_type_enum(self.config))

        # parents
        g.set_renderer("parents", self.render_parents_for_grid)

        # view action links to final asset record
        def asset_url(asset, i):
            return self.request.route_url(
                f"{asset.asset_type}_assets.view", uuid=asset.uuid
            )

        g.add_action("view", icon="eye", url=asset_url)

    def render_parents_for_grid(self, asset, field, value):
        parents = [str(p.parent) for p in asset._parents]
        return ", ".join(parents)

    def grid_row_class(self, asset, data, i):
        """ """
        if asset.archived:
            return "has-background-warning"
        return None


class AssetTypeMasterView(WuttaFarmMasterView):
    """
    Base class for "Asset Type" master views.

    A bit of a misnmer perhaps, this is *not* for the actual AssetType
    model, but rather the "secondary" types, e.g.  AnimalType,
    LandType etc.
    """


class AssetMasterView(WuttaFarmMasterView):
    """
    Base class for Asset master views
    """

    sort_defaults = "asset_name"

    filter_defaults = {
        "asset_name": {"active": True, "verb": "contains"},
        "archived": {"active": True, "verb": "is_false"},
    }

    def get_fallback_templates(self, template):
        templates = super().get_fallback_templates(template)

        if self.viewing:
            templates.insert(0, "/assets/master/view.mako")

        return templates

    def get_query(self, session=None):
        """ """
        model = self.app.model
        model_class = self.get_model_class()
        session = session or self.Session()
        return session.query(model_class).join(model.Asset)

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        model = self.app.model

        # thumbnail
        g.set_renderer("thumbnail", self.render_grid_thumbnail)
        g.set_label("thumbnail", "", column_only=True)
        g.set_centered("thumbnail")

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)
        g.set_sorter("drupal_id", model.Asset.drupal_id)
        g.set_filter("drupal_id", model.Asset.drupal_id)

        # asset_name
        g.set_link("asset_name")
        g.set_sorter("asset_name", model.Asset.asset_name)
        g.set_filter("asset_name", model.Asset.asset_name)

        # parents
        g.set_renderer("parents", self.render_parents_for_grid)

        # archived
        g.set_renderer("archived", "boolean")
        g.set_sorter("archived", model.Asset.archived)
        g.set_filter("archived", model.Asset.archived)

    def render_parents_for_grid(self, asset, field, value):
        parents = [str(p.parent) for p in asset.asset._parents]
        return ", ".join(parents)

    def grid_row_class(self, asset, data, i):
        """ """
        if asset.archived:
            return "has-background-warning"
        return None

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        asset = form.model_instance

        # asset_type
        if self.creating:
            f.remove("asset_type")
        else:
            f.set_node(
                "asset_type",
                WuttaDictEnum(self.request, get_asset_type_enum(self.config)),
            )
            f.set_readonly("asset_type")

        # parents
        if self.creating or self.editing:
            f.remove("parents")  # TODO: add support for this
        else:
            f.set_node("parents", AssetParentRefs(self.request))
            f.set_default("parents", [p.parent_uuid for p in asset.asset._parents])

        # notes
        f.set_widget("notes", "notes")

        # thumbnail_url
        if self.creating or self.editing:
            f.remove("thumbnail_url")

        # image_url
        if self.creating or self.editing:
            f.remove("image_url")

        # thumbnail
        if self.creating or self.editing:
            f.remove("thumbnail")
        elif asset.thumbnail_url:
            f.set_widget("thumbnail", ImageWidget("animal thumbnail"))
            f.set_default("thumbnail", asset.thumbnail_url)

        # image
        if self.creating or self.editing:
            f.remove("image")
        elif asset.image_url:
            f.set_widget("image", ImageWidget("animal image"))
            f.set_default("image", asset.image_url)

    def objectify(self, form):
        asset = super().objectify(form)

        if self.creating:
            model_class = self.get_model_class()
            asset.asset_type = model_class.__wutta_hint__["farmos_asset_type"]

        return asset

    def get_farmos_url(self, asset):
        return self.app.get_farmos_url(f"/asset/{asset.drupal_id}")

    def get_xref_buttons(self, asset):
        buttons = super().get_xref_buttons(asset)

        if asset.farmos_uuid:

            # TODO
            route = None
            if asset.asset_type == "animal":
                route = "farmos_animals.view"
            elif asset.asset_type == "group":
                route = "farmos_groups.view"
            elif asset.asset_type == "land":
                route = "farmos_land_assets.view"
            elif asset.asset_type == "structure":
                route = "farmos_structures.view"

            if route:
                buttons.append(
                    self.make_button(
                        "View farmOS record",
                        primary=True,
                        url=self.request.route_url(route, uuid=asset.farmos_uuid),
                        icon_left="eye",
                    )
                )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    AssetView = kwargs.get("AssetView", base["AssetView"])
    AssetView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for Users
"""

import colander

from wuttaweb.views import users as base


class UserView(base.UserView):
    """
    Custom master view for Users.
    """

    labels = {
        "farmos_uuid": "farmOS UUID",
        "drupal_id": "Drupal ID",
    }

    def get_template_context(self, context):
        context = super().get_template_context(context)

        if self.listing:
            context["farmos_refurl"] = self.app.get_farmos_url("/admin/people")

        return context

    def configure_form(self, form):
        """ """
        f = form
        super().configure_form(f)
        user = f.model_instance

        # farmos_uuid
        if not self.creating:
            f.fields.append("farmos_uuid")
            f.set_default("farmos_uuid", user.farmos_uuid or colander.null)

        # drupal_id
        if not self.creating:
            f.fields.append("drupal_id")
            f.set_default("drupal_id", user.drupal_id or colander.null)

    def get_xref_buttons(self, user):
        buttons = []

        if user.drupal_id:
            buttons.append(
                self.make_button(
                    "View in farmOS",
                    primary=True,
                    url=self.app.get_farmos_url(f"/user/{user.drupal_id}"),
                    target="_blank",
                    icon_left="external-link-alt",
                )
            )

        if user.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_users.view", uuid=user.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    local = globals()
    UserView = kwargs.get("UserView", local["UserView"])
    base.defaults(config, **{"UserView": UserView})


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaFarm data models
"""

# bring in all of wutta
from wuttjamaican.db.model import *

# wutta model extensions
from .users import WuttaFarmUser

# wuttafarm proper models
from .assets import AssetType, Asset, AssetParent
from .land import LandType, LandAsset
from .structures import StructureType, StructureAsset
from .animals import AnimalType, AnimalAsset
from .groups import GroupAsset
from .logs import LogType, ActivityLog



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Misc. utilities for web app
"""


def save_farmos_oauth2_token(request, token):
    """
    Common logic for saving the given OAuth2 token within the user
    session.  This function is called from 2 places:

    * :meth:`wuttafarm.web.views.auth.farmos_oauth_callback()`
    * :meth:`wuttafarm.web.views.farmos.master.FarmOSMasterView.get_farmos_client()`
    """
    # nb. we pretend the token expires 1 minute early, to avoid edge
    # cases around token refresh
    token["expires_at"] -= 60

    # save token to user session
    request.session["farmos.oauth2.token"] = token
